// N.Kurz, EE, GSI, 15-Jun-2016
// N.Kurz, EE, GSI, 04-Mar-2021: works for: RIO3 LynxOS 3.1
//                                          RIO4 LynxoS 4.0
//                                          RIO4 linux kernel version 2.6.24
//                                          RIO4 linux kernel version 2.6.33
// N.Kurz, EE, GSI, 09-Mar-2021: works for: IPV  linux kernel version 3.3.3
// N.Kurz, EE, GSI, 26-Mar-2021: works for: IFC  linux kernel version 4.1.8
// N.Kurz, EE, GSI, 01-Sep-2021: adopted (from MDPP) to VMMR
// N.Kurz, EE, GSI, 02-Sep-2021: common VMMR and MDPP readout
// N.Kurz, EE, GSI, 10-May-2022: added v775, v785 and v830 readout
// N.Kurz, EE, GSI, 24-Aug-2022: added VETAR, MADC, MTDC, VULOM scaler readout                       
//                               VETAR readout only possibel for RIO4 Linux, IFC and IPV
// B.Loeher, R3B, GSI, Nov-2022: support for TrLoII (VULOM trigger logic)

// read notes in "user change area". take care about different setup.usf for supported platforms

// N.Kurz, EE, GSI, 26-Mar-24: added RIO2 lynxos. vetar is readout with direct mapping, but without kernel module
//                             rio3 seems not to work with vetar.
//                             problem is not working cs/scr mapping with find_controller  


//                          ATTENTION: only RIO4 linux 2.6.33, IFC linux and RIO2 LynxOS tested.

//**********************************************************************************************************//

//---- begin user change area ------------------------------------------------------------------------------//

#define WR_TIME_STAMP  // white rabbit latched time stamp readout with VETAR

#undef  USE_TRLOII

#define N_VMMR         0
#define VMMR_OFF       0x20000

#define N_MDPP         0
#define MDPP_OFF       0x20000

#define N_V775         0
#define N_V775_CHA     32
#define V775_OFF       0x20000

#define N_V785         0
#define N_V785_CHA     32
#define V785_OFF       0x20000

#define N_V830         0
#define N_V830_CHA     32
#define V830_OFF       0x20000

#define N_MADC         0
#define N_MADC_CHA     32
#define MADC_OFF       0x20000

#define N_MTDC         0
#define N_MTDC_CHA     32
#define MTDC_OFF       0x20000

#define N_VUL_SCAL     0           // only one VULOM scaler can be used if no SIS3820 scaler is used !
#define N_VUL_SCAL_CHA 32
#define VUL_SCAL_OFF   0x1000000 

#define N_SIS3820      1           // only one SIS3820 scaler can be used if no VULOM scaler is used !
#define N_SIS3820_CHA  32
#define SIS3820_MASK   0xffffffff  // each '1' bit corresponds to an active channel. e.g. 0x0000f000 activates (and reads) only channels 12,13,14,15
#define SIS3820_OFF    0x1000000 


#define ADDR_OFF_VMMR     0x10000  
#define ADDR_OFF_MDPP     0x100000
#define ADDR_OFF_V775     0x200000
#define ADDR_OFF_V785     0x300000
#define ADDR_OFF_V830     0x400000
#define ADDR_OFF_MADC     0x500000
#define ADDR_OFF_MTDC     0x600000
#define ADDR_OFF_VUL_SCAL 0x1000000 // VULOM scaler has the same address offset as SIS3820 scaler. Only one scaler (either SIS or VULOM) is supported!
#define ADDR_OFF_SIS3820  0x1000000 // SIS3820 scaler has the same address offset as VULOM scaler. Only one scaler (either SIS or VULOM) is supported!

#define STATISTIC      1000000
#define MAX_TRIG_TYPE       16

// VME_BASE

// in case of multiple VMMR modules readout, VMMR VME base addresses must be consecutive with 
// a granularity VMMR_OFF

// VME_BASE has  to  be set for RIO4 LynxOS and RIO4 Linux kernel version 2.6.24
// VME_BASE has  to  be set for IPV
// VME_BASE must not be set for RIO2 - and RIO3 LynxOS and RIO4 Linux kernel version 2.6.33

// for RIO2 LynxOS set in setup.usf: LOC_MEM_BASE (1.crate) to VME base of first VMMR + 0xE7000000
//                                   LOC_MEM_LEN  (1.crate) to N_VMMR * VMMR_OFF             

// for RIO3 LynxOS set in setup.usf: LOC_MEM_BASE (1.crate) to VME base of first VMMR + 0x57000000
//                                   LOC_MEM_LEN  (1.crate) to N_VMMR * VMMR_OFF             

// for RIO4 Linux 2.6.33 set in setup.usf: LOC_MEM_BASE (1.crate) to VME base of first VMMR
//                                         LOC_MEM_LEN  (1.crate) to N_VMMR * VMMR_OFF            

#if (defined (RIO4) && ! defined (OSV_2_6_33)) || defined IPV || defined IFC
 #define VME_BASE  0x7000000
 #define VME_SIZE  0x2000000
#endif

#ifdef WR_TIME_STAMP
 #define SUB_SYSTEM_ID      0x600
 #define TS__ID_L16         0x3e1
 #define TS__ID_M16         0x4e1
 #define TS__ID_H16         0x5e1
 #define TS__ID_X16         0x6e1
#endif

//---- end user change area --------------------------------------------------------------------------------//

//**********************************************************************************************************//

#include <stdio.h>
#include <stdlib.h>
#include <s_veshe.h>
#include <sbs_def.h>

#ifdef IPV
 #include <sys/types.h>
 #include <sys/mman.h>
 #include <pevioctl.h>
 #include <pevulib.h>
#endif // IPV

#ifdef IFC
 #include <sys/types.h>
 #include <sys/mman.h>
 #include <sys/stat.h>
 #include <fcntl.h>
 #include <stdint.h>
 #include <errno.h>
 #include <f_user_exit.h>
 #include <altmasioctl.h>
 #include <altulib.h>

 //#define DEBUG_SLOW 1

#endif

#define RON  "\x1B[7m"
#define RES  "\x1B[0m"

#ifdef WR_TIME_STAMP
 #define WR_USE_TLU_DIRECT   1
 #define WR_RELEASE_FALLOUT  1
 #define VETAR_REGS_ADDR   0x50000000ULL
 #define VETAR_REGS_SIZE   0x400000
  #if defined (RIO4) && defined (OSV_2_6_33)
  #include <ces/xpc_vme.h>
  #include <ces/xpc_pci.h>
  #include <ces/CesXpcBridge.h>
  #include <ces/CesXpcBridge_Pci.h>
  #define SERIALIZE_IO __asm__ volatile ("eieio")
 #endif
 #ifdef RIO2
  #ifndef WR_USE_TLU_DIRECT
   #define WR_USE_TLU_DIRECT
  #endif
  #define WR_USE_TLU_DIRECT_WITHOUT_KERNEL_MODULE
  #define SERIALIZE_IO __asm__ volatile ("eieio")
 #endif
#endif

#if defined USE_TRLOII
 #include "trloii/trlo.h"
 #define ADDR_OFF_VULOM4 9
 struct TRLOIIModule trlo;
#endif

#if defined (RIO4) && defined (Lynx) 
 #define RIO4_DIRECT_VME_MAP 0xF0000000
#endif

#if defined (RIO4) && defined (Linux) && defined (OSV_2_6_24) 
 #define RIO4_DIRECT_VME_MAP 0xA0000000
#endif

#ifdef IPV
 // it seems that a32 elb mapping cannot handle d16 data sizing, therefore: 
 void f_ipv_a32_vme_mas_map ();
 void f_ipv_a32_vme_mas_free ();
 struct pev_node         *pev;
 struct pev_ioctl_map_pg  vme_mas_map;
 void                    *shm_vme_addr;
 //short                   *p16_vmmr_vmmr_vme_a32_base;
#endif // IPV

#ifdef IFC
 int triva_fd;
 const char* triva_driverhandle = "/dev/trigvme";
void f_ifc_a32_vme_mas_map  (int,   int, struct alt_ioctl_map_win*);
void f_ifc_a32_vme_mas_free (char*,      struct alt_ioctl_map_win*);
 int alt_fd;
 static struct alt_ioctl_map_win alt_win_modules;
#endif // IFC

CHARS      c_modnam[] = "f_user.c";
CHARS      c_line[256];

static short volatile *p16_virt_vme_base;
static short volatile *p16_virt_vme_mod_base;

static unsigned long  l_tr_ct[MAX_TRIG_TYPE];
static int            l_i;
static int            l_first = 0, l_first2 = 0, l_first3 = 0, l_first4 = 0;

//------------------------------------- VETAR---------------------------------------

#ifdef WR_TIME_STAMP
 #ifndef WR_USE_TLU_DIRECT_WITHOUT_KERNEL_MODULE
  #include <etherbone.h>
  #include <gsi_tm_latch.h> // wishbone devices
 #else
  typedef ADDR64 uint64_t;
  typedef uint64_t eb_data_t;

  #define VETAR_CRCSR_ADDR 0x280000
  // configbase CRCSR for slot 5
  #define VETAR_CRCSR_SIZE 0x80000
  /* size of cr/csr space*/
  #define VETAR_CTRL_ADDR  0x1400   
  // 5 * 0x400
  #define VETAR_CTRL_SIZE  0xA0
  #define PAGE_SHIFT       0x1000
  #define VETAR_VENDOR_ID  0x80031

  /* VETAR CR/CSR offsets: */
  #define VME_VENDOR_ID_OFFSET 0x24
  #define BOARD_ID             0x33
  #define REVISION_ID          0x43
  #define PROG_ID              0x7F

  #define FUN0ADER    0x7FF63
  #define FUN1ADER    0x7FF73
  #define INT_LEVEL   0x7ff5b
  #define INTVECTOR   0x7ff5f
  #define WB_32_64    0x7ff33
  #define BIT_SET_REG 0x7FFFB
  #define BIT_CLR_REG 0x7FFF7
  #define TIME        0x7FF3F
  #define BYTES       0x7FF37

  /* VETAR CR/CSR control values: */
  #define WB32        1
  #define WB64        0
  #define RESET_CORE  0x80
  #define ENABLE_CORE 0x10

  /* VME WB Interface*/
  #define CTRL               16
  #define MASTER_CTRL        24
  #define MASTER_ADD         32
  #define MASTER_DATA        40
  #define EMUL_DAT_WD        48
  #define WINDOW_OFFSET_LOW  56
  #define WINDOW_OFFSET_HIGH 64

  #define DIRECT_ACCESS_CONTROL 4

  // from gsi_tm_latch.h
  //register offsets
  #define GSI_TM_LATCH_FIFO_READY      0x000   // n..0 channel(n) timestamp(s) ready       (ro)
  #define GSI_TM_LATCH_FIFO_CLEAR      0x004   // n..0 channel(n) FIFO clear               (wo)
  #define GSI_TM_LATCH_TEST_CHANNELS   0x008   // Generate a test Event                    (wo)
  #define GSI_TM_LATCH_TRIG_ARMSTAT    0x00C   // n..0 channel(n) trigger armed status     (ro)
  #define GSI_TM_LATCH_TRIG_ARMSET     0x010   // n..0 channel(n) trigger set armed        (wo)
  #define GSI_TM_LATCH_TRIG_ARMCLR     0x014   // n..0 channel(n) trigger clr armed        (wo)
  #define GSI_TM_LATCH_TRIG_EDGESTAT   0x018   // n..0 channel(n) trigger edge status      (ro)
  #define GSI_TM_LATCH_TRIG_EDGEPOS    0x01C   // n..0 channel(n) trigger edge set pos     (wo)
  #define GSI_TM_LATCH_TRIG_EDGENEG    0x020   // n..0 channel(n) trigger edge set neg     (wo)
  // Channels Related Parameters
  #define GSI_TM_LATCH_CHNS_TOTAL      0x034   // Total Number of Channels in Device       (ro)
  #define GSI_TM_LATCH_CHNS_FIFOSIZE   0x038   // Total size of FIFOs                      (ro)
  // Channel to be selected n....0 and the other operations depend on selected channel
  #define GSI_TM_LATCH_CH_SELECT       0x058   // Channel Select                           (rw)
  // *IMP* All operations below depend on the Channel Selection
  #define GSI_TM_LATCH_FIFO_POP        0x05C   // pop the topmost FIFO Q Element           (wo)
  #define GSI_TM_LATCH_FIFO_TEST       0x060   // Generate a test Event Pulse              (wo)
  // pop just adjusts the pointer to the FIFO, it does not re-write a default value
  #define GSI_TM_LATCH_FIFO_CNT        0x064   // FIFO Queue   fill count                  (ro)
  #define GSI_TM_LATCH_FIFO_FTSHI      0x068   // timestamp HIGH words in cycles           (ro)
  #define GSI_TM_LATCH_FIFO_FTSLO      0x06c   // timestamp LOW words in cycles            (ro)
  #define GSI_TM_LATCH_FIFO_FTSSUB     0x070   // timestamp sub-cycle                      (ro)
 #endif

 #define WR_DEVICE_NAME "dev/wbm0"     // vetar vme
 #define WR_TLU_FIFO_NR       3        // vetar vme 

 void f_wr_init ();
 void f_wr_reset_tlu_fifo ();
 void f_vetar_init ();
 void f_a32_vme_mas_map_TLU ();
 void f_a32_vme_mas_free_TLU ();

 #ifdef WR_USE_TLU_DIRECT_WITHOUT_KERNEL_MODULE
  void f_a32_vme_mas_free_TLU ();
  void f_a24_vme_mas_free_CONTROL ();
  void f_a32_vme_mas_free_CRCSR ();
 #endif

 static FILE* dactl_handle = 0;

 #ifdef WR_RELEASE_FALLOUT
  static int tlu_address = 0x2000200; // fallout
 #else
  static int tlu_address = 0x2000100; // enigma
 #endif

 #ifdef RIO2
  #include <ces/vmelib.h>
  int   l_vme_perm;
  int   l_vme_addr;
  char *p_mem;

  struct pdparam_master param;
  unsigned long vmeaddr,vme_len,vme_am;
  unsigned long offset,size;
 #endif 

 #ifdef WR_USE_TLU_DIRECT
  // registers in the a32 region:
  static INTS4* ch_select;  
  static INTS4* ch_fifosize;
  static INTS4* fifoclear; 
  static INTS4* armset; 

  static INTS4* fifo_ready;
  static INTS4* fifo_cnt;
  static INTS4* ft_shi;
  static INTS4* ft_slo;
  static INTS4* ft_ssub;
  static INTS4* fifo_pop;

  #ifdef WR_USE_TLU_DIRECT_WITHOUT_KERNEL_MODULE
  // registers in the crcsr space:
  static INTS4* cr_vendor_id;
  static INTS4* cr_bit_set;
  static INTS4* cr_bit_clear;
  static INTS4* cr_wb_32_64;
  static INTS4* cr_fun0_ader;
  static INTS4* cr_fun1_ader;

  // registers for a24 control space:
  static INTS4* ctrl_master_control;
  static INTS4* ctrl_master_data;
  static INTS4* ctrl_emul_dat_wd;
  static INTS4* ctrl_window_offset_low;
  static INTS4* ctrl_direct_access_control;

  static char* p_TLU;
  static INTU4* pl_virt_vme_base_tlu;
  static INTU4* pl_virt_vme_base_crcsr;
  static INTU4* pl_virt_vme_base_control;
  #endif
 
  #if defined (RIO4) && defined (OSV_2_6_33)
   static CesUint64            ll_TLU;
   static unsigned long        p_TLU;
   static struct CesXpcBridge *lXpcBridge;
   static CesUint32            l_vme_size;
   static CesUint64            ll_vme_base;
  #endif
  #ifdef IFC
   int           alt_fd;
   static struct alt_ioctl_map_win  alt_win_vetar;
   static INTU4 volatile           *p32_virt_vme_base;
   static char*   p_TLU;
   static INTU4*  pl_virt_vme_base_tlu;
  #endif // IFC
 #endif // WR_USE_TLU_DIRECT

 static  long               l_eb_first1 = 0, l_eb_first2 = 0;
 #ifndef  WR_USE_TLU_DIRECT
  static  eb_status_t       eb_stat;
  static  eb_device_t       eb_device;
  static  eb_socket_t       eb_socket;
  static  eb_cycle_t        eb_cycle;
  static  struct sdb_device sdbDevice; 
  static  eb_address_t      wrTLU;
  static  eb_address_t      wrTLUFIFO;
  static  int               nDevices;
 #endif // WR_USE_TLU_DIRECT 
 static  long               l_used_tlu_fifo = 1 << WR_TLU_FIFO_NR;    
 static  eb_data_t          eb_fifo_cha;
 static  eb_data_t          eb_fifo_size;
 static  eb_data_t          eb_tlu_high_ts; // high 32 bit 
 static  eb_data_t          eb_tlu_low_ts;  // low  32 bit

 static  eb_data_t          eb_tlu_high_ts_prev; // high 32 bit
 static  eb_data_t          eb_tlu_low_ts_prev;  // low  32 bit

 static  eb_data_t          eb_stat_before; // status before etherbone cycle
 static  eb_data_t          eb_stat_after;  // status after  etherbone cycle
 static  eb_data_t          eb_fifo_ct_brd; // TLU FIFO fill counter before TLU read
 static  eb_data_t          eb_fifo_ct_ard; // TLU FIFO fill counter after  TLU read

 static  unsigned long long ll_ts_hi;
 static  unsigned long long ll_ts_lo;
 static  unsigned long long ll_ts;
 static  unsigned long long ll_x16;
 static  unsigned long long ll_h16;
 static  unsigned long long ll_m16;
 static  unsigned long long ll_l16;

 static  unsigned long long ll_timestamp;
 static  unsigned long long ll_actu_timestamp;
 static  unsigned long long ll_prev_timestamp;
 static  unsigned long long ll_diff_timestamp;

 static INTU4  l_time_l16;
 static INTU4  l_time_m16;
 static INTU4  l_time_h16;
 static INTU4  l_time_x16;

 static long l_check_wr_err = 0;
 static long l_wr_init_ct   = 0;
 static long l_err_wr_ct    = 0;

#endif // WR_TIME_STAMP

//------------------------------------- VMMR---------------------------------------

static unsigned long l_vmmr_base_16[N_VMMR];
static unsigned long l_vmmr_base_32[N_VMMR];

static INTU4 volatile *pl_vmmr_dat [N_VMMR];

static short volatile *p16_vmmr_0x6000_addr_src    [N_VMMR];
static short volatile *p16_vmmr_0x6002_addr_reg    [N_VMMR];
static short volatile *p16_vmmr_0x6004_mod_id      [N_VMMR];
static short volatile *p16_vmmr_0x6008_soft_reset  [N_VMMR];
static short volatile *p16_vmmr_0x600E_firm_rev    [N_VMMR];

static short volatile *p16_vmmr_0x6030_buf_dat_len [N_VMMR];
static short volatile *p16_vmmr_0x6032_dat_len_form[N_VMMR];
static short volatile *p16_vmmr_0x6034_read_reset  [N_VMMR];
static short volatile *p16_vmmr_0x6036_mult_evt    [N_VMMR];
static short volatile *p16_vmmr_0x6038_mark_typ    [N_VMMR];
static short volatile *p16_vmmr_0x603A_start_acq   [N_VMMR];
static short volatile *p16_vmmr_0x603C_fifo_reset  [N_VMMR];
static short volatile *p16_vmmr_0x603E_dat_ready   [N_VMMR];

static short volatile *p16_vmmr_0x6042_active_buses[N_VMMR];
static short volatile *p16_vmmr_0x6046_timing_resol[N_VMMR];

static short volatile *p16_vmmr_0x6050_win_start   [N_VMMR];
static short volatile *p16_vmmr_0x6054_win_width   [N_VMMR];
static short volatile *p16_vmmr_0x6058_ext_trig_src[N_VMMR];
static short volatile *p16_vmmr_0x605A_mmr_trig_src[N_VMMR];
static short volatile *p16_vmmr_0x605C_out_trig_src[N_VMMR];
static short volatile *p16_vmmr_0x605E_trig_out    [N_VMMR];

static short volatile *p16_vmmr_0x6068_nim3_io     [N_VMMR];
static short volatile *p16_vmmr_0x606A_nim2_io     [N_VMMR];
static short volatile *p16_vmmr_0x606C_nim1_io     [N_VMMR];
static short volatile *p16_vmmr_0x606E_nim0_io     [N_VMMR];

static short volatile *p16_vmmr_0x6070_pulser_stat [N_VMMR];
static short volatile *p16_vmmr_0x6072_pulser_ampl [N_VMMR];

static short volatile *p16_vmmr_0x6090_reset_ctr_ab[N_VMMR];
static short volatile *p16_vmmr_0x6096_ts_sources  [N_VMMR];
static short volatile *p16_vmmr_0x6098_ts_divisor  [N_VMMR];

static short volatile *p16_vmmr_0x6100_select_bus  [N_VMMR];
static short volatile *p16_vmmr_0x6110_fe_addr_wr  [N_VMMR];
static short volatile *p16_vmmr_0x6112_fe_data_wr  [N_VMMR];
static short volatile *p16_vmmr_0x611A_bus_delay   [N_VMMR];

static short volatile *p16_vmmr_0x631C_no_docu     [N_VMMR];
static short volatile *p16_vmmr_0x6320_no_docu     [N_VMMR];


static INTU4 volatile *p32_vmmr_tmp;
static short volatile *p16_vmmr_tmp;

static INTU4 volatile *p32_vmmr_virt_vme_base; 
static short volatile *p16_vmmr_virt_vme_base;

//------------------------------------- MDPP ---------------------------------------

static unsigned long l_mdpp_base_16[N_MDPP];
static unsigned long l_mdpp_base_32[N_MDPP];

static INTU4 volatile *pl_mdpp_dat [N_MDPP];

static short volatile *p16_mdpp_0x6000_addr_src    [N_MDPP];
static short volatile *p16_mdpp_0x6002_addr_reg    [N_MDPP];
static short volatile *p16_mdpp_0x6004_mod_id      [N_MDPP];
static short volatile *p16_mdpp_0x6008_soft_reset  [N_MDPP];
static short volatile *p16_mdpp_0x600E_firm_rev    [N_MDPP];

static short volatile *p16_mdpp_0x6030_buf_dat_len [N_MDPP];
static short volatile *p16_mdpp_0x6032_dat_len_form[N_MDPP];
static short volatile *p16_mdpp_0x6034_read_reset  [N_MDPP];
static short volatile *p16_mdpp_0x6036_mult_evt    [N_MDPP];
static short volatile *p16_mdpp_0x6038_mark_typ    [N_MDPP];
static short volatile *p16_mdpp_0x603A_start_acq   [N_MDPP];
static short volatile *p16_mdpp_0x603C_fifo_reset  [N_MDPP];
static short volatile *p16_mdpp_0x603E_dat_ready   [N_MDPP];

static short volatile *p16_mdpp_0x6042_tdc_resol   [N_MDPP];
static short volatile *p16_mdpp_0x6044_out_form    [N_MDPP];
static short volatile *p16_mdpp_0x6046_adc_resol   [N_MDPP];

static short volatile *p16_mdpp_0x6050_win_start   [N_MDPP];
static short volatile *p16_mdpp_0x6054_win_width   [N_MDPP];
static short volatile *p16_mdpp_0x6058_trig_src    [N_MDPP];
static short volatile *p16_mdpp_0x605A_trig_src_2  [N_MDPP];
static short volatile *p16_mdpp_0x605C_first_hit   [N_MDPP];
static short volatile *p16_mdpp_0x605E_trig_out    [N_MDPP];

static short volatile *p16_mdpp_0x6090_reset_ctr_ab[N_MDPP];

// mdpp channel settings
static short volatile *p16_mdpp_0x6100_sel_cha_pair[N_MDPP];
static short volatile *p16_mdpp_0x6110_tf_int_diff [N_MDPP];
static short volatile *p16_mdpp_0x611A_gain        [N_MDPP];
static short volatile *p16_mdpp_0x611C_thresh0     [N_MDPP];
static short volatile *p16_mdpp_0x611E_thresh1     [N_MDPP];
static short volatile *p16_mdpp_0x6124_shaping_time[N_MDPP];
static short volatile *p16_mdpp_0x6126_blr         [N_MDPP];
static short volatile *p16_mdpp_0x6128_reset_time  [N_MDPP];
static short volatile *p16_mdpp_0x612A_sig_ris_time[N_MDPP];
static short volatile *p16_mdpp_0x6112_pz0         [N_MDPP];
static short volatile *p16_mdpp_0x6114_pz1         [N_MDPP];
//

static INTU4 volatile *p32_mdpp_tmp;
static short volatile *p16_mdpp_tmp;

static INTU4 volatile *p32_mdpp_virt_vme_base;  
static short volatile *p16_mdpp_virt_vme_base;

//------------------------------------- V775 ---------------------------------------

static unsigned long l_v775_base_16[N_V775];
static unsigned long l_v775_base_32[N_V775];

static INTU4 volatile *pl_v775_dat [N_V775];

static short volatile *p16_v775_0x1000_firm_rev         [N_V775];
static short volatile *p16_v775_0x1002_geo_addr         [N_V775];
static short volatile *p16_v775_0x1006_bit_set_1        [N_V775];
static short volatile *p16_v775_0x1008_bit_clear_1      [N_V775];
static short volatile *p16_v775_0x100e_status_1         [N_V775];
static short volatile *p16_v775_0x1010_control          [N_V775];
static short volatile *p16_v775_0x1032_bit_set_2        [N_V775];
static short volatile *p16_v775_0x1034_bit_clear_2      [N_V775];
static short volatile *p16_v775_0x1040_evt_ctr_reset    [N_V775];
static short volatile *p16_v775_0x1060_full_scale_range [N_V775];
static short volatile *p16_v775_threshholds             [N_V775][N_V775_CHA]; // 0x1080 -> 0x10be

static INTU4 volatile *p32_v775_tmp;
static short volatile *p16_v775_tmp;

static INTU4 volatile *p32_v775_virt_vme_base; 
static short volatile *p16_v775_virt_vme_base;

unsigned long l_v775_head;

//------------------------------------- V785 ---------------------------------------

static unsigned long l_v785_base_16[N_V785];
static unsigned long l_v785_base_32[N_V785];

static INTU4 volatile *pl_v785_dat [N_V785];

static short volatile *p16_v785_0x1000_firm_rev         [N_V785];
static short volatile *p16_v785_0x1002_geo_addr         [N_V785];
static short volatile *p16_v785_0x1006_bit_set_1        [N_V785];
static short volatile *p16_v785_0x1008_bit_clear_1      [N_V785];
static short volatile *p16_v785_0x100e_status_1         [N_V785];
static short volatile *p16_v785_0x1010_control          [N_V785];
static short volatile *p16_v785_0x1032_bit_set_2        [N_V785];
static short volatile *p16_v785_0x1034_bit_clear_2      [N_V785];
static short volatile *p16_v785_0x1040_evt_ctr_reset    [N_V785];
static short volatile *p16_v785_threshholds             [N_V785][N_V785_CHA]; // 0x1080 -> 0x10be

static INTU4 volatile *p32_v785_tmp;
static short volatile *p16_v785_tmp;

static INTU4 volatile *p32_v785_virt_vme_base; 
static short volatile *p16_v785_virt_vme_base;

unsigned long l_v785_head;

//------------------------------------- V830 ---------------------------------------

static unsigned long l_v830_base_16[N_V830];
static unsigned long l_v830_base_32[N_V830];

static INTU4 volatile *pl_v830_dat [N_V830];

static INTU4 volatile *p32_v830_0x1100_ena_chan  [N_V830];
static short volatile *p16_v830_0x1108_control   [N_V830];
static short volatile *p16_v830_0x110A_bit_set   [N_V830];
static short volatile *p16_v830_0x110C_bit_clear [N_V830];
static short volatile *p16_v830_0x110E_status    [N_V830];
static short volatile *p16_v830_0x1110_geo_addr  [N_V830];
static short volatile *p16_v830_0x1120_soft_reset[N_V830];
static short volatile *p16_v830_0x1132_firm_rev  [N_V830];

static INTU4 volatile *p32_v830_tmp;
static short volatile *p16_v830_tmp;

static INTU4 volatile *p32_v830_virt_vme_base; 
static short volatile *p16_v830_virt_vme_base;

unsigned long l_v830_head;

//------------------------------------- MADC ---------------------------------------

static unsigned long l_madc_base_16[N_MADC];
static unsigned long l_madc_base_32[N_MADC];

static INTU4 volatile *pl_madc_dat [N_MADC];

static short volatile *p16_madc_0x6000_addr_src    [N_MADC];
static short volatile *p16_madc_0x6002_addr_reg    [N_MADC];
static short volatile *p16_madc_0x6004_mod_id      [N_MADC];
static short volatile *p16_madc_0x6008_soft_reset  [N_MADC];
static short volatile *p16_madc_0x600E_firm_rev    [N_MADC];

static short volatile *p16_madc_0x6030_buf_dat_len [N_MADC];
static short volatile *p16_madc_0x6032_dat_len_form[N_MADC];
static short volatile *p16_madc_0x6034_read_reset  [N_MADC];
static short volatile *p16_madc_0x6036_mult_evt    [N_MADC];
static short volatile *p16_madc_0x6038_mark_typ    [N_MADC];
static short volatile *p16_madc_0x603A_start_acq   [N_MADC];
static short volatile *p16_madc_0x603C_fifo_reset  [N_MADC];
static short volatile *p16_madc_0x603E_dat_ready   [N_MADC];

static short volatile *p16_madc_0x6040_bank_op     [N_MADC];
static short volatile *p16_madc_0x6042_adc_resol   [N_MADC];
static short volatile *p16_madc_0x6044_out_form    [N_MADC];
static short volatile *p16_madc_0x6046_adc_overwr  [N_MADC];
static short volatile *p16_madc_0x6048_slc_off     [N_MADC];
static short volatile *p16_madc_0x604A_skip_oor    [N_MADC];

static short volatile *p16_madc_0x6050_hold_delay0 [N_MADC];
static short volatile *p16_madc_0x6052_hold_delay1 [N_MADC];
static short volatile *p16_madc_0x6054_hold_width0 [N_MADC];
static short volatile *p16_madc_0x6056_hold_width1 [N_MADC];
static short volatile *p16_madc_0x6058_use_gg      [N_MADC];

static short volatile *p16_madc_0x6060_input_range [N_MADC];
static short volatile *p16_madc_0x6062_ecl_term    [N_MADC];
static short volatile *p16_madc_0x6064_ecl_gat1_osc[N_MADC];
static short volatile *p16_madc_0x6066_ecl_fs_reset[N_MADC];
static short volatile *p16_madc_0x6068_ecl_busy    [N_MADC];
static short volatile *p16_madc_0x606A_nim_gat1_osc[N_MADC];
static short volatile *p16_madc_0x606C_nim_fc_reset[N_MADC];
static short volatile *p16_madc_0x606E_nim_busy    [N_MADC];

static short volatile *p16_madc_0x6070_pulse_stat  [N_MADC];
static short volatile *p16_madc_0x6090_reset_ctr_ab[N_MADC];

static short volatile *p16_madc_thresholds         [N_MADC][N_MADC_CHA];

static INTU4 volatile *p32_madc_tmp;
static short volatile *p16_madc_tmp;

static INTU4 volatile *p32_madc_virt_vme_base; 
static short volatile *p16_madc_virt_vme_base;

unsigned long l_madc_head;

//------------------------------------- MTDC ---------------------------------------

static unsigned long l_mtdc_base_16[N_MTDC];
static unsigned long l_mtdc_base_32[N_MTDC];

static INTU4 volatile *pl_mtdc_dat [N_MTDC];

static short volatile *p16_mtdc_0x6000_addr_src        [N_MTDC];
static short volatile *p16_mtdc_0x6002_addr_reg        [N_MTDC];
static short volatile *p16_mtdc_0x6004_mod_id          [N_MTDC];
static short volatile *p16_mtdc_0x6008_soft_reset      [N_MTDC];
static short volatile *p16_mtdc_0x600E_firm_rev        [N_MTDC];

static short volatile *p16_mtdc_0x6030_buf_dat_len     [N_MTDC];
static short volatile *p16_mtdc_0x6032_dat_len_form    [N_MTDC];
static short volatile *p16_mtdc_0x6034_read_reset      [N_MTDC];
static short volatile *p16_mtdc_0x6036_mult_evt        [N_MTDC];
static short volatile *p16_mtdc_0x6038_mark_typ        [N_MTDC];
static short volatile *p16_mtdc_0x603A_start_acq       [N_MTDC];
static short volatile *p16_mtdc_0x603C_fifo_reset      [N_MTDC];
static short volatile *p16_mtdc_0x603E_dat_ready       [N_MTDC];

static short volatile *p16_mtdc_0x6040_bank_op         [N_MTDC];
static short volatile *p16_mtdc_0x6042_tdc_resol       [N_MTDC];
static short volatile *p16_mtdc_0x6044_out_form        [N_MTDC];

static short volatile *p16_mtdc_0x6050_bank0_win_start [N_MTDC];
static short volatile *p16_mtdc_0x6052_bank1_win_start [N_MTDC];
static short volatile *p16_mtdc_0x6054_bank0_win_width [N_MTDC];
static short volatile *p16_mtdc_0x6056_bank1_win_width [N_MTDC];
static short volatile *p16_mtdc_0x6058_bank0_trig_src  [N_MTDC];
static short volatile *p16_mtdc_0x605A_bank1_trig_src  [N_MTDC];
static short volatile *p16_mtdc_0x605C_first_hit       [N_MTDC];

static short volatile *p16_mtdc_0x6060_negative_edge   [N_MTDC];
static short volatile *p16_mtdc_0x6062_ecl_term        [N_MTDC];
static short volatile *p16_mtdc_0x6064_ecl_trig1_osc   [N_MTDC];
static short volatile *p16_mtdc_0x6066_ecl_out_config  [N_MTDC];
static short volatile *p16_mtdc_0x6068_trig_select     [N_MTDC];
static short volatile *p16_mtdc_0x606A_nim_trig1_osc   [N_MTDC];
static short volatile *p16_mtdc_0x606E_nim_busy        [N_MTDC];

static short volatile *p16_mtdc_0x6070_pulse_stat      [N_MTDC];
static short volatile *p16_mtdc_0x6090_reset_ctr_ab    [N_MTDC];

static INTU4 volatile *p32_mtdc_tmp;
static short volatile *p16_mtdc_tmp;

static INTU4 volatile *p32_mtdc_virt_vme_base; 
static short volatile *p16_mtdc_virt_vme_base;

unsigned long l_mtdc_head;

//------------------------------------- VUL_SCAL -----------------------------------

static unsigned long   l_vul_scal_base_32[N_VUL_SCAL];

static INTU4 volatile *p32_vul_scal_0x04_latch    [N_VUL_SCAL];
static INTU4 volatile *p32_vul_scal_0x08_reset    [N_VUL_SCAL];
static INTU4 volatile *p32_vul_scal_0x0C_inv_cha  [N_VUL_SCAL];
static INTU4 volatile *p32_vul_scal_0x10_ena_cha  [N_VUL_SCAL];
static INTU4 volatile *p32_vul_scal_0x84_fr_scal  [N_VUL_SCAL]; // internal counter fed by fixed clock
                                                                // to make rates
static INTU4 volatile *p32_vul_scal_0x88_ena_cha_r[N_VUL_SCAL]; // read back enable pattern
static INTU4 volatile *p32_vul_scal_0x8C_inv_cha_r[N_VUL_SCAL]; // read back invert pattern
static INTU4 volatile *p32_vul_scal_dat           [N_VUL_SCAL][N_VUL_SCAL_CHA];

static INTU4 volatile *p32_vul_scal_tmp;
static INTU4 volatile *p32_vul_scal_virt_vme_base; 

//------------------------------------- SIS3820 -----------------------------------

static unsigned long   l_sis3820_base_32[N_SIS3820];
static unsigned long   l_sis3820_channels[N_SIS3820];

static INTU4 volatile *p32_sis3820_0x000_control_status          [N_SIS3820]; 
static INTU4 volatile *p32_sis3820_0x100_operation_mode          [N_SIS3820];
static INTU4 volatile *p32_sis3820_0x104_copy_disable            [N_SIS3820];
static INTU4 volatile *p32_sis3820_0x400_key_reset               [N_SIS3820];
static INTU4 volatile *p32_sis3820_0x404_key_sdram_fifo_reset    [N_SIS3820];
static INTU4 volatile *p32_sis3820_0x418_key_operation_enable    [N_SIS3820];
static INTU4 volatile *p32_sis3820_0x800_counter_shadow          [N_SIS3820][N_SIS3820_CHA];
static INTU4 volatile *p32_sis3820_0x800000_fifo_addr            [N_SIS3820];

static INTU4 volatile *p32_sis3820_tmp;
static INTU4 volatile *p32_sis3820_virt_vme_base; 

//----------------------------------------------------------------------------------

static INTU4 volatile *pl_tmp;  
static INTU4 volatile *pl_dat_save;         
static long  l_ct;

/**************************************************************************/

int f_user_get_virt_ptr (INTS4  *pl_loc_hwacc, INTS4  pl_rem_cam[])
{
  long l_i, l_j;

  //----------------------------------- VETAR --------------------------------------
  
  #ifdef WR_TIME_STAMP
  if (l_eb_first1 == 0)
  {
    l_eb_first1 = 1;
  
    #ifdef WR_USE_TLU_DIRECT
    #ifndef WR_USE_TLU_DIRECT_WITHOUT_KERNEL_MODULE
    printm ("switch VETAR to direct access TLU mode \n");
    
    dactl_handle= fopen ("/sys/class/vetar/vetar0/dactl", "a");
    if(dactl_handle==NULL)
    {
      printm ("ERROR>> Could not open dactl control sysfs handle, exiting.. \n");
      exit(-1); // probably wrong driver?
    }
    else
    {
      fprintf(dactl_handle,"%d",tlu_address);
      fclose(dactl_handle);
    }
    printm ("f_user_get_virt_ptr for DIRECT TLU \n");
    #else // WR_USE_TLU_DIRECT_WITHOUT_KERNEL_MODULE
    f_vetar_init ();
    // this will switch on vetar addresses    
    f_a32_vme_mas_map_TLU ();
    p_TLU = (char*) pl_virt_vme_base_tlu;
    printm ("mapped TLU base address 0x%x to virtual address 0x%08lx\n", VETAR_REGS_ADDR, (long) p_TLU);
    
    // used in init function
    ch_select   = (INTS4*) ((char*) (p_TLU) + GSI_TM_LATCH_CH_SELECT);
    ch_fifosize = (INTS4*) ((char*) (p_TLU) + GSI_TM_LATCH_CHNS_FIFOSIZE);
    fifoclear   = (INTS4*) ((char*) (p_TLU) + GSI_TM_LATCH_FIFO_CLEAR);
    armset      = (INTS4*) ((char*) (p_TLU) + GSI_TM_LATCH_TRIG_ARMSET);
    
    // set here pointers to mapped registers used in readout function:  
    fifo_ready  = (INTS4*) ((char*) (p_TLU) + GSI_TM_LATCH_FIFO_READY);
    fifo_cnt    = (INTS4*) ((char*) (p_TLU) + GSI_TM_LATCH_FIFO_CNT);
    ft_shi      = (INTS4*) ((char*) (p_TLU) + GSI_TM_LATCH_FIFO_FTSHI);
    ft_slo      = (INTS4*) ((char*) (p_TLU) + GSI_TM_LATCH_FIFO_FTSLO);
    ft_ssub     = (INTS4*) ((char*) (p_TLU) + GSI_TM_LATCH_FIFO_FTSSUB);
    fifo_pop    = (INTS4*) ((char*) (p_TLU) + GSI_TM_LATCH_FIFO_POP);
    #endif // WR_USE_TLU_DIRECT_WITHOUT_KERNEL_MODULE
    
    if (l_first2 == 0)
    {
      l_first2 = 1;
      #if defined (RIO4) && defined (OSV_2_6_33)
      //printm ("f_user_get_virt_ptr for DIRECT TLU \n");
      l_vme_size  = (CesUint32)VETAR_REGS_SIZE;
      ll_vme_base = (CesUint64)VETAR_REGS_ADDR; // for slot number 5, old mapping

      lXpcBridge = CesXpcBridge_GetByName ("VME Bridge");
      if (!lXpcBridge)
      {
        printm ("ERROR>> CesXpcBridge_GetByName(\"%s\") failed, exiting \n", "VME Bridge");
        exit (-1);
      }

      ll_TLU = (CesUint64) CesXpcBridge_MasterMapVirt64
                                              (lXpcBridge, ll_vme_base, l_vme_size, XPC_VME_A32_STD_USER);
      if (ll_TLU == 0xffffffffffffffffULL)
      {
        printm ("ERROR>> CesXpcBridge_MasterMap(%s, %x, %x, %x) failed, exiting \n", "VME Bridge", 
                                                           ll_vme_base, l_vme_size, XPC_VME_A32_STD_USER);
        exit (-1);
      }
      printm ("TLU - VTRANS VME mapping: (A32 / AM =: 9) \n");
      printm ("mapped VME address 0x%08lx \n",(long)  ll_vme_base);
      printm ("         with size 0x%08x \n", l_vme_size);
      printm ("map.bus_address    0x%08lx \n", (long) ll_TLU);

      p_TLU = (unsigned long) ll_TLU; // supress warnings on 32bit systems
      #endif      

      #ifdef IFC
      f_ifc_a32_vme_mas_map (VETAR_REGS_ADDR, VETAR_REGS_SIZE, &alt_win_vetar);
      p_TLU = (char*) p16_virt_vme_base;
      printm ("mapped TLU base address 0x%lx to virtual address 0x%08lx\n",VETAR_REGS_ADDR, (long)  p_TLU);
      #endif
    }  

    // used in init function
    ch_select   	= (INTS4*) ((char*)(p_TLU) 	+ GSI_TM_LATCH_CH_SELECT);
    ch_fifosize 	= (INTS4*) ((char*)(p_TLU) 	+ GSI_TM_LATCH_CHNS_FIFOSIZE);
    fifoclear 		= (INTS4*) ((char*)(p_TLU) 	+ GSI_TM_LATCH_FIFO_CLEAR);
    armset 		= (INTS4*) ((char*)(p_TLU) 	+ GSI_TM_LATCH_TRIG_ARMSET);
    
    // set here pointers to mapped registers used in readout function:  
    fifo_ready		= (INTS4*) ((char*)(p_TLU) 	+ GSI_TM_LATCH_FIFO_READY);
    fifo_cnt 		= (INTS4*) ((char*)(p_TLU) 	+ GSI_TM_LATCH_FIFO_CNT);
    ft_shi 		= (INTS4*) ((char*)(p_TLU) 	+ GSI_TM_LATCH_FIFO_FTSHI);
    ft_slo 		= (INTS4*) ((char*)(p_TLU) 	+ GSI_TM_LATCH_FIFO_FTSLO);
    ft_ssub		= (INTS4*) ((char*)(p_TLU) 	+ GSI_TM_LATCH_FIFO_FTSSUB);
    fifo_pop		= (INTS4*) ((char*)(p_TLU) 	+ GSI_TM_LATCH_FIFO_POP);
  
    #else  // WR_USE_TLU_DIRECT
    
    printm ("f_user_get_virt_ptr  switches VETAR to etherbone access mode... \n");
    dactl_handle= fopen ("/sys/class/vetar/vetar0/dactl", "a");
    if(dactl_handle==NULL)
    {
      printm ("!!! Could not open dactl control sysfs handle!!! \n");
      exit(-1); // probably wrong driver?
    }
    else
    {
      //fprintf(dactl_handle,"%d",tlu_direct_off);
      fprintf(dactl_handle,"4294967295"); // -1 or hex 0xffffffff
      fclose(dactl_handle);
      //sleep(1);      
    }
    
    if ((eb_stat = eb_socket_open(EB_ABI_CODE, 0, EB_ADDR32|EB_DATA32, &eb_socket)) != EB_OK)
    {
      printm (RON"ERROR>>"RES" etherbone eb_open_socket, status: %s \n", eb_status(eb_stat));
    }

    if ((eb_stat = eb_device_open(eb_socket, WR_DEVICE_NAME, EB_ADDR32|EB_DATA32, 3, &eb_device)) != EB_OK)
    {
      printm (RON"ERROR>>"RES" etherbone eb_device_open, status: %s \n", eb_status(eb_stat));
    }

    nDevices = 1;
    if ((eb_stat = eb_sdb_find_by_identity(eb_device, GSI_TM_LATCH_VENDOR,
                                                  GSI_TM_LATCH_PRODUCT, &sdbDevice, &nDevices)) != EB_OK)
    {
      printm (RON"ERROR>>"RES" etherbone TLU eb_sdb_find_by_identity, status: %s \n", eb_status(eb_stat));
    }
  
    if (nDevices == 0)
    {
      printm (RON"ERROR>>"RES" etherbone no TLU found, status: %s \n", eb_status(eb_stat));
    }

    if (nDevices > 1)
    {
      printm (RON"ERROR>>"RES" etherbone too many TLUsfound, status: %s \n", eb_status(eb_stat));
    }

    /* Record the address of the device */
    wrTLU     = sdbDevice.sdb_component.addr_first;
    //wrTLUFIFO = wrTLU + GSI_TM_LATCH_FIFO_OFFSET + WR_TLU_FIFO_NR * GSI_TM_LATCH_FIFO_INCR;
    #endif //WR_USE_TLU_DIRECT
  }
  #endif // WR_TIME_STAMP
  
  //------------------------------Mesytec, CAEN and other modules mapping -----------

  //#if defined RIO4 && defined (OSV_2_6_33)
  #if (defined RIO4 && defined (OSV_2_6_33)) || defined RIO2   
  p32_vmmr_virt_vme_base     = (INTU4*) ((INTU4)pl_loc_hwacc + (INTU4)ADDR_OFF_VMMR);
  p16_vmmr_virt_vme_base     = (short*) ((INTU4)pl_loc_hwacc + (INTU4)ADDR_OFF_VMMR);
  p32_mdpp_virt_vme_base     = (INTU4*) ((INTU4)pl_loc_hwacc + (INTU4)ADDR_OFF_MDPP);
  p16_mdpp_virt_vme_base     = (short*) ((INTU4)pl_loc_hwacc + (INTU4)ADDR_OFF_MDPP);
  p32_v775_virt_vme_base     = (INTU4*) ((INTU4)pl_loc_hwacc + (INTU4)ADDR_OFF_V775);
  p16_v775_virt_vme_base     = (short*) ((INTU4)pl_loc_hwacc + (INTU4)ADDR_OFF_V775);
  p32_v785_virt_vme_base     = (INTU4*) ((INTU4)pl_loc_hwacc + (INTU4)ADDR_OFF_V785);
  p16_v785_virt_vme_base     = (short*) ((INTU4)pl_loc_hwacc + (INTU4)ADDR_OFF_V785);
  p32_v830_virt_vme_base     = (INTU4*) ((INTU4)pl_loc_hwacc + (INTU4)ADDR_OFF_V830);
  p16_v830_virt_vme_base     = (short*) ((INTU4)pl_loc_hwacc + (INTU4)ADDR_OFF_V830);
  p32_madc_virt_vme_base     = (INTU4*) ((INTU4)pl_loc_hwacc + (INTU4)ADDR_OFF_MADC);
  p16_madc_virt_vme_base     = (short*) ((INTU4)pl_loc_hwacc + (INTU4)ADDR_OFF_MADC);
  p32_mtdc_virt_vme_base     = (INTU4*) ((INTU4)pl_loc_hwacc + (INTU4)ADDR_OFF_MTDC);
  p16_mtdc_virt_vme_base     = (short*) ((INTU4)pl_loc_hwacc + (INTU4)ADDR_OFF_MTDC);
  p32_vul_scal_virt_vme_base = (INTU4*) ((INTU4)pl_loc_hwacc + (INTU4)ADDR_OFF_VUL_SCAL);  
  p32_sis3820_virt_vme_base  = (INTU4*) ((INTU4)pl_loc_hwacc + (INTU4)ADDR_OFF_SIS3820 );  
  #endif

  #if defined RIO4 && ! defined (OSV_2_6_33)
  p32_vmmr_virt_vme_base = (INTU4*)((INTU4)RIO4_DIRECT_VME_MAP + (INTU4)VME_BASE);
  p16_vmmr_virt_vme_base = (short*)p32_vmmr_virt_vme_base;
  #endif 

  #ifdef IPV
  f_map_ipv_a32_elb_vme (&p32_vmmr_virt_vme_base);
  p32_vmmr_virt_vme_base  = (INTU4*) ((INTU4)p32_vmmr_virt_vme_base + (INTU4)VME_BASE);
  f_ipv_a32_vme_mas_map ();
  #endif // IPV

  #ifdef IFC
  if (l_first3 == 0)
  {
    l_first3 = 1; 
    f_ifc_a32_vme_mas_map (VME_BASE, VME_SIZE, &alt_win_modules);
    p16_virt_vme_mod_base = p16_virt_vme_base;
    
    p32_vmmr_virt_vme_base     = pl_loc_hwacc          + (ADDR_OFF_VMMR >> 2);
    p16_vmmr_virt_vme_base     = p16_virt_vme_mod_base + (ADDR_OFF_VMMR >> 1);
    p32_mdpp_virt_vme_base     = pl_loc_hwacc          + (ADDR_OFF_MDPP >> 2);
    p16_mdpp_virt_vme_base     = p16_virt_vme_mod_base + (ADDR_OFF_MDPP >> 1);
    p32_v775_virt_vme_base     = pl_loc_hwacc          + (ADDR_OFF_V775 >> 2);
    p16_v775_virt_vme_base     = p16_virt_vme_mod_base + (ADDR_OFF_V775 >> 1);
    p32_v785_virt_vme_base     = pl_loc_hwacc          + (ADDR_OFF_V785 >> 2);
    p16_v785_virt_vme_base     = p16_virt_vme_mod_base + (ADDR_OFF_V785 >> 1);
    p32_v830_virt_vme_base     = pl_loc_hwacc          + (ADDR_OFF_V830 >> 2);
    p16_v830_virt_vme_base     = p16_virt_vme_mod_base + (ADDR_OFF_V830 >> 1);
    p32_madc_virt_vme_base     = pl_loc_hwacc          + (ADDR_OFF_MADC >> 2);
    p16_madc_virt_vme_base     = p16_virt_vme_mod_base + (ADDR_OFF_MADC >> 1);
    p32_mtdc_virt_vme_base     = pl_loc_hwacc          + (ADDR_OFF_MTDC >> 2);
    p16_mtdc_virt_vme_base     = p16_virt_vme_mod_base + (ADDR_OFF_MTDC >> 1);
    p32_vul_scal_virt_vme_base = pl_loc_hwacc          + (ADDR_OFF_VUL_SCAL >> 2);
    p32_sis3820_virt_vme_base  = pl_loc_hwacc          + (ADDR_OFF_VUL_SCAL >> 2);
  }  
  #endif // IFC


  if (l_first4 == 0)
  {
    l_first4 = 1;
  
    //------------------------------------- VMMR ---------------------------------------

    for (l_i=0; l_i<N_VMMR; l_i++)
    {
      l_vmmr_base_32  [l_i] = (long)   p32_vmmr_virt_vme_base + (long)(VMMR_OFF * l_i);
      pl_vmmr_dat     [l_i] = (INTU4*) (l_vmmr_base_32[l_i] + 0x0);

      l_vmmr_base_16  [l_i] = (long)   p16_vmmr_virt_vme_base + (long)(VMMR_OFF * l_i);  
      p16_vmmr_0x6000_addr_src    [l_i] = (short*) (l_vmmr_base_16[l_i] + 0x6000); 
      p16_vmmr_0x6002_addr_reg    [l_i] = (short*) (l_vmmr_base_16[l_i] + 0x6002);
      p16_vmmr_0x6004_mod_id      [l_i] = (short*) (l_vmmr_base_16[l_i] + 0x6004);
      p16_vmmr_0x6008_soft_reset  [l_i] = (short*) (l_vmmr_base_16[l_i] + 0x6008);
      p16_vmmr_0x600E_firm_rev    [l_i] = (short*) (l_vmmr_base_16[l_i] + 0x600E);

      p16_vmmr_0x6030_buf_dat_len [l_i] = (short*) (l_vmmr_base_16[l_i] + 0x6030);
      p16_vmmr_0x6032_dat_len_form[l_i] = (short*) (l_vmmr_base_16[l_i] + 0x6032);
      p16_vmmr_0x6034_read_reset  [l_i] = (short*) (l_vmmr_base_16[l_i] + 0x6034);
      p16_vmmr_0x6036_mult_evt    [l_i] = (short*) (l_vmmr_base_16[l_i] + 0x6036);
      p16_vmmr_0x6038_mark_typ    [l_i] = (short*) (l_vmmr_base_16[l_i] + 0x6038);
      p16_vmmr_0x603A_start_acq   [l_i] = (short*) (l_vmmr_base_16[l_i] + 0x603A);
      p16_vmmr_0x603C_fifo_reset  [l_i] = (short*) (l_vmmr_base_16[l_i] + 0x603C);
      p16_vmmr_0x603E_dat_ready   [l_i] = (short*) (l_vmmr_base_16[l_i] + 0x603E);

      p16_vmmr_0x6042_active_buses[l_i] = (short*) (l_vmmr_base_16[l_i] + 0x6042);
      p16_vmmr_0x6046_timing_resol[l_i] = (short*) (l_vmmr_base_16[l_i] + 0x6046);

      p16_vmmr_0x6050_win_start   [l_i] = (short*) (l_vmmr_base_16[l_i] + 0x6050);
      p16_vmmr_0x6054_win_width   [l_i] = (short*) (l_vmmr_base_16[l_i] + 0x6054);
      p16_vmmr_0x6058_ext_trig_src[l_i] = (short*) (l_vmmr_base_16[l_i] + 0x6058);
      p16_vmmr_0x605A_mmr_trig_src[l_i] = (short*) (l_vmmr_base_16[l_i] + 0x605A);
      p16_vmmr_0x605C_out_trig_src[l_i] = (short*) (l_vmmr_base_16[l_i] + 0x605C);
      p16_vmmr_0x605E_trig_out    [l_i] = (short*) (l_vmmr_base_16[l_i] + 0x605E);

      p16_vmmr_0x6068_nim3_io     [l_i] = (short*) (l_vmmr_base_16[l_i] + 0x6068);
      p16_vmmr_0x606A_nim2_io     [l_i] = (short*) (l_vmmr_base_16[l_i] + 0x606A);
      p16_vmmr_0x606C_nim1_io     [l_i] = (short*) (l_vmmr_base_16[l_i] + 0x606C);
      p16_vmmr_0x606E_nim0_io     [l_i] = (short*) (l_vmmr_base_16[l_i] + 0x606E);      

      p16_vmmr_0x6070_pulser_stat [l_i] = (short*) (l_vmmr_base_16[l_i] + 0x6070);
      p16_vmmr_0x6072_pulser_ampl [l_i] = (short*) (l_vmmr_base_16[l_i] + 0x6072);

      p16_vmmr_0x6090_reset_ctr_ab[l_i] = (short*) (l_vmmr_base_16[l_i] + 0x6090);
      p16_vmmr_0x6096_ts_sources  [l_i] = (short*) (l_vmmr_base_16[l_i] + 0x6096);
      p16_vmmr_0x6098_ts_divisor  [l_i] = (short*) (l_vmmr_base_16[l_i] + 0x6098);

      p16_vmmr_0x6100_select_bus  [l_i] = (short*) (l_vmmr_base_16[l_i] + 0x6100);
      p16_vmmr_0x6110_fe_addr_wr  [l_i] = (short*) (l_vmmr_base_16[l_i] + 0x6110);
      p16_vmmr_0x6112_fe_data_wr  [l_i] = (short*) (l_vmmr_base_16[l_i] + 0x6112);
      p16_vmmr_0x611A_bus_delay   [l_i] = (short*) (l_vmmr_base_16[l_i] + 0x611A);

      p16_vmmr_0x631C_no_docu     [l_i] = (short*) (l_vmmr_base_16[l_i] + 0x631C);
      p16_vmmr_0x6320_no_docu     [l_i] = (short*) (l_vmmr_base_16[l_i] + 0x6320);      
    }

    //------------------------------------- MDPP ---------------------------------------

    for (l_i=0; l_i<N_MDPP; l_i++)
    {
      l_mdpp_base_32  [l_i] = (long)   p32_mdpp_virt_vme_base + (long)(MDPP_OFF * l_i);
      pl_mdpp_dat     [l_i] = (INTU4*) (l_mdpp_base_32[l_i] + 0x0);

      l_mdpp_base_16  [l_i] = (long)   p16_mdpp_virt_vme_base + (long)(MDPP_OFF * l_i);  
      p16_mdpp_0x6000_addr_src    [l_i] = (short*) (l_mdpp_base_16[l_i] + 0x6000); 
      p16_mdpp_0x6002_addr_reg    [l_i] = (short*) (l_mdpp_base_16[l_i] + 0x6002);
      p16_mdpp_0x6004_mod_id      [l_i] = (short*) (l_mdpp_base_16[l_i] + 0x6004);
      p16_mdpp_0x6008_soft_reset  [l_i] = (short*) (l_mdpp_base_16[l_i] + 0x6008);
      p16_mdpp_0x600E_firm_rev    [l_i] = (short*) (l_mdpp_base_16[l_i] + 0x600E);

      p16_mdpp_0x6030_buf_dat_len [l_i] = (short*) (l_mdpp_base_16[l_i] + 0x6030);
      p16_mdpp_0x6032_dat_len_form[l_i] = (short*) (l_mdpp_base_16[l_i] + 0x6032);
      p16_mdpp_0x6034_read_reset  [l_i] = (short*) (l_mdpp_base_16[l_i] + 0x6034);
      p16_mdpp_0x6036_mult_evt    [l_i] = (short*) (l_mdpp_base_16[l_i] + 0x6036);
      p16_mdpp_0x6038_mark_typ    [l_i] = (short*) (l_mdpp_base_16[l_i] + 0x6038);
      p16_mdpp_0x603A_start_acq   [l_i] = (short*) (l_mdpp_base_16[l_i] + 0x603A);
      p16_mdpp_0x603C_fifo_reset  [l_i] = (short*) (l_mdpp_base_16[l_i] + 0x603C);
      p16_mdpp_0x603E_dat_ready   [l_i] = (short*) (l_mdpp_base_16[l_i] + 0x603E);

      p16_mdpp_0x6042_tdc_resol   [l_i] = (short*) (l_mdpp_base_16[l_i] + 0x6042);
      p16_mdpp_0x6044_out_form    [l_i] = (short*) (l_mdpp_base_16[l_i] + 0x6044);
      p16_mdpp_0x6046_adc_resol   [l_i] = (short*) (l_mdpp_base_16[l_i] + 0x6046);

      p16_mdpp_0x6050_win_start   [l_i] = (short*) (l_mdpp_base_16[l_i] + 0x6050);
      p16_mdpp_0x6054_win_width   [l_i] = (short*) (l_mdpp_base_16[l_i] + 0x6054);
      p16_mdpp_0x6058_trig_src    [l_i] = (short*) (l_mdpp_base_16[l_i] + 0x6058);
      p16_mdpp_0x605A_trig_src_2  [l_i] = (short*) (l_mdpp_base_16[l_i] + 0x605A);
      p16_mdpp_0x605C_first_hit   [l_i] = (short*) (l_mdpp_base_16[l_i] + 0x605C);
      p16_mdpp_0x605E_trig_out    [l_i] = (short*) (l_mdpp_base_16[l_i] + 0x605E);

      p16_mdpp_0x6100_sel_cha_pair[l_i] = (short*) (l_mdpp_base_16[l_i] + 0x6100);
      p16_mdpp_0x6110_tf_int_diff [l_i] = (short*) (l_mdpp_base_16[l_i] + 0x6110);
      p16_mdpp_0x611A_gain        [l_i] = (short*) (l_mdpp_base_16[l_i] + 0x611A);
      p16_mdpp_0x611C_thresh0     [l_i] = (short*) (l_mdpp_base_16[l_i] + 0x611C);
      p16_mdpp_0x611E_thresh1     [l_i] = (short*) (l_mdpp_base_16[l_i] + 0x611E);
      p16_mdpp_0x6124_shaping_time[l_i] = (short*) (l_mdpp_base_16[l_i] + 0x6124);
      p16_mdpp_0x6126_blr         [l_i] = (short*) (l_mdpp_base_16[l_i] + 0x6126);
      p16_mdpp_0x6128_reset_time  [l_i] = (short*) (l_mdpp_base_16[l_i] + 0x6128);
      p16_mdpp_0x612A_sig_ris_time[l_i] = (short*) (l_mdpp_base_16[l_i] + 0x612A);
      p16_mdpp_0x6112_pz0         [l_i] = (short*) (l_mdpp_base_16[l_i] + 0x6112);
      p16_mdpp_0x6114_pz1         [l_i] = (short*) (l_mdpp_base_16[l_i] + 0x6114);

      p16_mdpp_0x6090_reset_ctr_ab[l_i] = (short*) (l_mdpp_base_16[l_i] + 0x6090);
    }


    //------------------------------------- V775 ---------------------------------------

    for (l_i=0; l_i<N_V775; l_i++)
    {
      l_v775_base_32  [l_i] = (long)   p32_v775_virt_vme_base + (long)(V775_OFF * l_i);
      pl_v775_dat     [l_i] = (INTU4*) (l_v775_base_32[l_i] + 0x0);

      l_v775_base_16  [l_i] = (long)   p16_v775_virt_vme_base + (long)(V775_OFF * l_i);

      p16_v775_0x1000_firm_rev         [l_i] = (short*) (l_v775_base_16[l_i] + 0x1000);  
      p16_v775_0x1002_geo_addr         [l_i] = (short*) (l_v775_base_16[l_i] + 0x1002); 
      p16_v775_0x1006_bit_set_1        [l_i] = (short*) (l_v775_base_16[l_i] + 0x1006); 
      p16_v775_0x1008_bit_clear_1      [l_i] = (short*) (l_v775_base_16[l_i] + 0x1008);
      p16_v775_0x100e_status_1         [l_i] = (short*) (l_v775_base_16[l_i] + 0x100e); 
      p16_v775_0x1010_control          [l_i] = (short*) (l_v775_base_16[l_i] + 0x1010); 
      p16_v775_0x1032_bit_set_2        [l_i] = (short*) (l_v775_base_16[l_i] + 0x1032); 
      p16_v775_0x1034_bit_clear_2      [l_i] = (short*) (l_v775_base_16[l_i] + 0x1034); 
      p16_v775_0x1040_evt_ctr_reset    [l_i] = (short*) (l_v775_base_16[l_i] + 0x1040); 
      p16_v775_0x1060_full_scale_range [l_i] = (short*) (l_v775_base_16[l_i] + 0x1060); 

      for (l_j=0; l_j<N_V775_CHA; l_j++)
      {
        p16_v775_threshholds [l_i][l_j] = (short*) (l_v775_base_16[l_i] + 0x1080) + 2*l_j;
      }
    }

    //------------------------------------- V785 ---------------------------------------

    for (l_i=0; l_i<N_V785; l_i++)
    {
      l_v785_base_32  [l_i] = (long)   p32_v785_virt_vme_base + (long)(V785_OFF * l_i);
      pl_v785_dat     [l_i] = (INTU4*) (l_v785_base_32[l_i] + 0x0);

      l_v785_base_16  [l_i] = (long)   p16_v785_virt_vme_base + (long)(V785_OFF * l_i);

      p16_v785_0x1000_firm_rev         [l_i] = (short*) (l_v785_base_16[l_i] + 0x1000);  
      p16_v785_0x1002_geo_addr         [l_i] = (short*) (l_v785_base_16[l_i] + 0x1002); 
      p16_v785_0x1006_bit_set_1        [l_i] = (short*) (l_v785_base_16[l_i] + 0x1006); 
      p16_v785_0x1008_bit_clear_1      [l_i] = (short*) (l_v785_base_16[l_i] + 0x1008);
      p16_v785_0x100e_status_1         [l_i] = (short*) (l_v785_base_16[l_i] + 0x100e);
      p16_v785_0x1010_control          [l_i] = (short*) (l_v785_base_16[l_i] + 0x1010); 
      p16_v785_0x1032_bit_set_2        [l_i] = (short*) (l_v785_base_16[l_i] + 0x1032); 
      p16_v785_0x1034_bit_clear_2      [l_i] = (short*) (l_v785_base_16[l_i] + 0x1034); 
      p16_v785_0x1040_evt_ctr_reset    [l_i] = (short*) (l_v785_base_16[l_i] + 0x1040); 

      for (l_j=0; l_j<N_V785_CHA; l_j++)
      {
        p16_v785_threshholds [l_i][l_j] = (short*) (l_v785_base_16[l_i] + 0x1080) + 2*l_j;
      }
    }

    //------------------------------------- V830 ---------------------------------------

    for (l_i=0; l_i<N_V830; l_i++)
    {
      l_v830_base_32  [l_i] = (long)   p32_v830_virt_vme_base + (long)(V830_OFF * l_i);
      pl_v830_dat     [l_i] = (INTU4*) (l_v830_base_32[l_i] + 0x0);

      l_v830_base_16  [l_i] = (long)   p16_v830_virt_vme_base + (long)(V830_OFF * l_i);

      p32_v830_0x1100_ena_chan  [l_i] = (INTU4*) (l_v830_base_16[l_i] + 0x1100); 
      p16_v830_0x1108_control   [l_i] = (short*) (l_v830_base_16[l_i] + 0x1108); 
      p16_v830_0x110A_bit_set   [l_i] = (short*) (l_v830_base_16[l_i] + 0x110a); 
      p16_v830_0x110C_bit_clear [l_i] = (short*) (l_v830_base_16[l_i] + 0x110c); 
      p16_v830_0x110E_status    [l_i] = (short*) (l_v830_base_16[l_i] + 0x110e); 
      p16_v830_0x1110_geo_addr  [l_i] = (short*) (l_v830_base_16[l_i] + 0x1110); 
      p16_v830_0x1120_soft_reset[l_i] = (short*) (l_v830_base_16[l_i] + 0x1120); 
      p16_v830_0x1132_firm_rev  [l_i] = (short*) (l_v830_base_16[l_i] + 0x1132); 
    }

    //------------------------------------- MADC ---------------------------------------

    for (l_i=0; l_i<N_MADC; l_i++)
    {
      l_madc_base_32  [l_i] = (long)   p32_madc_virt_vme_base + (long)(MADC_OFF * l_i);
      pl_madc_dat     [l_i] = (INTU4*) (l_madc_base_32[l_i] + 0x0);

      l_madc_base_16  [l_i] = (long)   p16_madc_virt_vme_base + (long)(MADC_OFF * l_i);

      p16_madc_0x6000_addr_src    [l_i] = (short*) (l_madc_base_16[l_i] + 0x6000); 
      p16_madc_0x6002_addr_reg    [l_i] = (short*) (l_madc_base_16[l_i] + 0x6002);
      p16_madc_0x6004_mod_id      [l_i] = (short*) (l_madc_base_16[l_i] + 0x6004);
      p16_madc_0x6008_soft_reset  [l_i] = (short*) (l_madc_base_16[l_i] + 0x6008);
      p16_madc_0x600E_firm_rev    [l_i] = (short*) (l_madc_base_16[l_i] + 0x600E);

      p16_madc_0x6030_buf_dat_len [l_i] = (short*) (l_madc_base_16[l_i] + 0x6030);
      p16_madc_0x6032_dat_len_form[l_i] = (short*) (l_madc_base_16[l_i] + 0x6032);
      p16_madc_0x6034_read_reset  [l_i] = (short*) (l_madc_base_16[l_i] + 0x6034);
      p16_madc_0x6036_mult_evt    [l_i] = (short*) (l_madc_base_16[l_i] + 0x6036);
      p16_madc_0x6038_mark_typ    [l_i] = (short*) (l_madc_base_16[l_i] + 0x6038);
      p16_madc_0x603A_start_acq   [l_i] = (short*) (l_madc_base_16[l_i] + 0x603A);
      p16_madc_0x603C_fifo_reset  [l_i] = (short*) (l_madc_base_16[l_i] + 0x603C);
      p16_madc_0x603E_dat_ready   [l_i] = (short*) (l_madc_base_16[l_i] + 0x603E);

      p16_madc_0x6040_bank_op     [l_i] = (short*) (l_madc_base_16[l_i] + 0x6040);
      p16_madc_0x6042_adc_resol   [l_i] = (short*) (l_madc_base_16[l_i] + 0x6042);
      p16_madc_0x6044_out_form    [l_i] = (short*) (l_madc_base_16[l_i] + 0x6044);
      p16_madc_0x6046_adc_overwr  [l_i] = (short*) (l_madc_base_16[l_i] + 0x6046);
      p16_madc_0x6048_slc_off     [l_i] = (short*) (l_madc_base_16[l_i] + 0x6048);
      p16_madc_0x604A_skip_oor    [l_i] = (short*) (l_madc_base_16[l_i] + 0x604A);

      p16_madc_0x6050_hold_delay0 [l_i] = (short*) (l_madc_base_16[l_i] + 0x6050);
      p16_madc_0x6052_hold_delay1 [l_i] = (short*) (l_madc_base_16[l_i] + 0x6052);
      p16_madc_0x6054_hold_width0 [l_i] = (short*) (l_madc_base_16[l_i] + 0x6054);
      p16_madc_0x6056_hold_width1 [l_i] = (short*) (l_madc_base_16[l_i] + 0x6056);
      p16_madc_0x6058_use_gg      [l_i] = (short*) (l_madc_base_16[l_i] + 0x6058);

      p16_madc_0x6060_input_range [l_i] = (short*) (l_madc_base_16[l_i] + 0x6060);
      p16_madc_0x6062_ecl_term    [l_i] = (short*) (l_madc_base_16[l_i] + 0x6062);
      p16_madc_0x6064_ecl_gat1_osc[l_i] = (short*) (l_madc_base_16[l_i] + 0x6064);
      p16_madc_0x6066_ecl_fs_reset[l_i] = (short*) (l_madc_base_16[l_i] + 0x6066);
      p16_madc_0x6068_ecl_busy    [l_i] = (short*) (l_madc_base_16[l_i] + 0x6068);
      p16_madc_0x606A_nim_gat1_osc[l_i] = (short*) (l_madc_base_16[l_i] + 0x606A);
      p16_madc_0x606C_nim_fc_reset[l_i] = (short*) (l_madc_base_16[l_i] + 0x606C);
      p16_madc_0x606E_nim_busy    [l_i] = (short*) (l_madc_base_16[l_i] + 0x606E);

      p16_madc_0x6070_pulse_stat  [l_i] = (short*) (l_madc_base_16[l_i] + 0x6070);
      p16_madc_0x6090_reset_ctr_ab[l_i] = (short*) (l_madc_base_16[l_i] + 0x6090);

      for (l_j=0; l_j<N_MADC_CHA; l_j++)
      {
        p16_madc_thresholds[l_i][l_j] = (short*) (l_madc_base_16[l_i] + 0x4000 + (2*l_j));
      }
    }

    //------------------------------------- MTDC ---------------------------------------

    for (l_i=0; l_i<N_MTDC; l_i++)
    {
      l_mtdc_base_32  [l_i] = (long)   p32_mtdc_virt_vme_base + (long)(MTDC_OFF * l_i);
      pl_mtdc_dat     [l_i] = (INTU4*) (l_mtdc_base_32[l_i] + 0x0);

      l_mtdc_base_16  [l_i] = (long)   p16_mtdc_virt_vme_base + (long)(MTDC_OFF * l_i);

      p16_mtdc_0x6000_addr_src       [l_i] = (short*) (l_mtdc_base_16[l_i] + 0x6000); 
      p16_mtdc_0x6002_addr_reg       [l_i] = (short*) (l_mtdc_base_16[l_i] + 0x6002);
      p16_mtdc_0x6004_mod_id         [l_i] = (short*) (l_mtdc_base_16[l_i] + 0x6004);
      p16_mtdc_0x6008_soft_reset     [l_i] = (short*) (l_mtdc_base_16[l_i] + 0x6008);
      p16_mtdc_0x600E_firm_rev       [l_i] = (short*) (l_mtdc_base_16[l_i] + 0x600E);

      p16_mtdc_0x6030_buf_dat_len    [l_i] = (short*) (l_mtdc_base_16[l_i] + 0x6030);
      p16_mtdc_0x6032_dat_len_form   [l_i] = (short*) (l_mtdc_base_16[l_i] + 0x6032);
      p16_mtdc_0x6034_read_reset     [l_i] = (short*) (l_mtdc_base_16[l_i] + 0x6034);
      p16_mtdc_0x6036_mult_evt       [l_i] = (short*) (l_mtdc_base_16[l_i] + 0x6036);
      p16_mtdc_0x6038_mark_typ       [l_i] = (short*) (l_mtdc_base_16[l_i] + 0x6038);
      p16_mtdc_0x603A_start_acq      [l_i] = (short*) (l_mtdc_base_16[l_i] + 0x603A);
      p16_mtdc_0x603C_fifo_reset     [l_i] = (short*) (l_mtdc_base_16[l_i] + 0x603C);
      p16_mtdc_0x603E_dat_ready      [l_i] = (short*) (l_mtdc_base_16[l_i] + 0x603E);

      p16_mtdc_0x6040_bank_op        [l_i] = (short*) (l_mtdc_base_16[l_i] + 0x6040);
      p16_mtdc_0x6042_tdc_resol      [l_i] = (short*) (l_mtdc_base_16[l_i] + 0x6042);
      p16_mtdc_0x6044_out_form       [l_i] = (short*) (l_mtdc_base_16[l_i] + 0x6044);

      p16_mtdc_0x6050_bank0_win_start[l_i] = (short*) (l_mtdc_base_16[l_i] + 0x6050);
      p16_mtdc_0x6052_bank1_win_start[l_i] = (short*) (l_mtdc_base_16[l_i] + 0x6052);
      p16_mtdc_0x6054_bank0_win_width[l_i] = (short*) (l_mtdc_base_16[l_i] + 0x6054);
      p16_mtdc_0x6056_bank1_win_width[l_i] = (short*) (l_mtdc_base_16[l_i] + 0x6056);
      p16_mtdc_0x6058_bank0_trig_src [l_i] = (short*) (l_mtdc_base_16[l_i] + 0x6058);
      p16_mtdc_0x605A_bank1_trig_src [l_i] = (short*) (l_mtdc_base_16[l_i] + 0x605A);
      p16_mtdc_0x605C_first_hit      [l_i] = (short*) (l_mtdc_base_16[l_i] + 0x605C);            

      p16_mtdc_0x6060_negative_edge  [l_i] = (short*) (l_mtdc_base_16[l_i] + 0x6060);
      p16_mtdc_0x6062_ecl_term       [l_i] = (short*) (l_mtdc_base_16[l_i] + 0x6062);
      p16_mtdc_0x6064_ecl_trig1_osc  [l_i] = (short*) (l_mtdc_base_16[l_i] + 0x6064);
      p16_mtdc_0x6066_ecl_out_config [l_i] = (short*) (l_mtdc_base_16[l_i] + 0x6066);
      p16_mtdc_0x6068_trig_select    [l_i] = (short*) (l_mtdc_base_16[l_i] + 0x6068);
      p16_mtdc_0x606A_nim_trig1_osc  [l_i] = (short*) (l_mtdc_base_16[l_i] + 0x606A);
      p16_mtdc_0x606E_nim_busy       [l_i] = (short*) (l_mtdc_base_16[l_i] + 0x606E);

      p16_mtdc_0x6070_pulse_stat     [l_i] = (short*) (l_mtdc_base_16[l_i] + 0x6070);
      p16_mtdc_0x6090_reset_ctr_ab   [l_i] = (short*) (l_mtdc_base_16[l_i] + 0x6090);
    }

    //------------------------------------- VUL_SCAL -----------------------------------

    for (l_i=0; l_i<N_VUL_SCAL; l_i++)
    {
      l_vul_scal_base_32  [l_i] = (long)   p32_vul_scal_virt_vme_base + (long)(VUL_SCAL_OFF * l_i);
      //pl_vul_scal_dat     [l_i] = (INTU4*) (l_vul_scal_base_32[l_i] + 0x0);

      p32_vul_scal_0x04_latch     [l_i] = (INTU4*) (l_vul_scal_base_32[l_i] + 0x04);
      p32_vul_scal_0x08_reset     [l_i] = (INTU4*) (l_vul_scal_base_32[l_i] + 0x08);
      p32_vul_scal_0x0C_inv_cha   [l_i] = (INTU4*) (l_vul_scal_base_32[l_i] + 0x0C);
      p32_vul_scal_0x10_ena_cha   [l_i] = (INTU4*) (l_vul_scal_base_32[l_i] + 0x10);
      p32_vul_scal_0x84_fr_scal   [l_i] = (INTU4*) (l_vul_scal_base_32[l_i] + 0x84);
      p32_vul_scal_0x88_ena_cha_r [l_i] = (INTU4*) (l_vul_scal_base_32[l_i] + 0x88);
      p32_vul_scal_0x8C_inv_cha_r [l_i] = (INTU4*) (l_vul_scal_base_32[l_i] + 0x8C);

      for (l_j=0; l_j<N_VUL_SCAL_CHA; l_j++)
      {
        p32_vul_scal_dat[l_i][l_j] = (INTU4*) (l_vul_scal_base_32[l_i] + 0x4 + (4*l_j));
        //printm ("scal ptr: %p \n", p32_vul_scal_dat[l_i][l_j]); 
      }
    }  
    
    //------------------------------------- SIS3820 -----------------------------------

    for (l_i=0; l_i<N_SIS3820; l_i++)
    {
      l_sis3820_base_32  [l_i] = (long)   p32_sis3820_virt_vme_base + (long)(SIS3820_OFF * l_i);

      p32_sis3820_0x000_control_status          [l_i] = (INTU4*) (l_sis3820_base_32[l_i] + 0x000); 
      p32_sis3820_0x100_operation_mode          [l_i] = (INTU4*) (l_sis3820_base_32[l_i] + 0x100);
      p32_sis3820_0x104_copy_disable            [l_i] = (INTU4*) (l_sis3820_base_32[l_i] + 0x104);
      p32_sis3820_0x400_key_reset               [l_i] = (INTU4*) (l_sis3820_base_32[l_i] + 0x400);
      p32_sis3820_0x404_key_sdram_fifo_reset    [l_i] = (INTU4*) (l_sis3820_base_32[l_i] + 0x404);
      p32_sis3820_0x418_key_operation_enable    [l_i] = (INTU4*) (l_sis3820_base_32[l_i] + 0x418);
      p32_sis3820_0x800000_fifo_addr            [l_i] = (INTU4*) (l_sis3820_base_32[l_i] + 0x800000);

      for (l_j=0; l_j<N_SIS3820_CHA; l_j++)
      {
        p32_sis3820_0x800_counter_shadow[l_i][l_j] = (INTU4*) (l_sis3820_base_32[l_i] + 0x800 + (4*l_j));
        //printm ("scal ptr: %p \n", p32_sis3820_0x800_counter_shadow[l_i][l_j]); 
      }


      // find number of active channels by counting the '1's in the enable bitmask
      l_sis3820_channels[l_i] = 0;
      for (l_j=0; l_j<N_SIS3820_CHA; l_j++) 
      {
        if ( 0x1 & (SIS3820_MASK>>l_j) ) 
        {
          l_sis3820_channels[l_i]++; 
        }
      }
      printm("sis3820 scaler has mask 0x%x and %d enabled channels\n", SIS3820_MASK, l_sis3820_channels[l_i]);
    }  
    
    //----------------------------------- TRLOII ---------------------------------------
    #if defined USE_TRLOII
      #if (defined RIO4 && defined (OSV_2_6_33)) || (defined IFC)
        trlo.trlo = trlo_setup_map_hardware(ADDR_OFF_VULOM4, &trlo._trlo_unmapinfo);
      #endif
    #endif
    //----------------------------------------------------------------------------------

  }
    
  return (0);

}
/*****************************************************************************/

int f_user_init (CHARU  bh_crate_nr,
                 INTS4 *pl_loc_hwacc,
                 INTS4 *pl_rem_cam,
                 INTS4 *pl_stat)
{
  INTU4 l_dat;
  int   l_i, l_j;

  #ifdef IFC
  #ifndef NOVMECONF
  int alt;
  struct alt_ioctl_vme_mas_conf mas_p;

  alt = alt_init();
 
  printm ("alt: %d 0x%x \n", alt, alt);
  if (alt == 0)
  {
    printm("ERROR>> cannot allocate data structures to control ALTHEA \n");
    exit( -1);
  }
  if (alt < 0)
  {
    printm("ERROR>> cannot find ALTHEA interface\n");
    exit( -1);
  }
  
  // the settings below are not really required, because they provoke
  // no perfomance improvements.

  mas_p.req_level = 3;
  mas_p.req_mode  = VME_REQ_ROR;
  if ((alt_vme_mas_set_conf (&mas_p)) == -1)
  {
    printm ("ERROR>> could not write  VME master configuration \n");
  }

  if ((alt_vme_mas_get_conf (&mas_p)) == -1)
  {
    printm ("ERROR>> could not read VME master configuration \n");
  }
  else
  {
    printm ("VME request level: %d", mas_p.req_level);
    printm ("VME request mode:  %d, VME_REQ_ROR: %d \n", mas_p.req_mode, VME_REQ_ROR);
  }
  #endif // NOVMECONF
  #endif // IFC

  if (l_first == 0)
  {
    l_first = 1;
    for (l_i=0; l_i<MAX_TRIG_TYPE; l_i++)
    {
      l_tr_ct[l_i] = 0;
    }
  }
  
  switch (bh_crate_nr)
  {
    case 0:
      
      //------------------------------------- VETAR --------------------------------------

      #ifdef WR_TIME_STAMP
      f_wr_init ();
      #endif
      
      //------------------------------------- VMMR ---------------------------------------
 
      // initalize VMMR 
      for (l_i=0; l_i<N_VMMR; l_i++)
      {
        //VMMR
        *p16_vmmr_0x6008_soft_reset  [l_i] = 1; sleep (1);
        *p16_vmmr_0x6004_mod_id      [l_i] = l_i;
        *p16_vmmr_0x6036_mult_evt    [l_i] = 0;    // single event mode
        *p16_vmmr_0x6038_mark_typ    [l_i] = 0;    // provide event counter, not time stamp
        *p16_vmmr_0x6032_dat_len_form[l_i] = 2;    // 32 bit data 

        *p16_vmmr_0x6042_active_buses[l_i] = 0xff; // activate all 8 mmr buses
        *p16_vmmr_0x6050_win_start   [l_i] = 3996; // 3996: 500 ns before trigger
        *p16_vmmr_0x6054_win_width   [l_i] = 200;  //  200: 1 us in total (500 ns before, 500 ns after trigger)
        *p16_vmmr_0x6058_ext_trig_src[l_i] = 1;    // nim TI input trigger
        *p16_vmmr_0x605A_mmr_trig_src[l_i] = 0;    // no vmmr trigger from mmr
        *p16_vmmr_0x605C_out_trig_src[l_i] = 0;

        *p16_vmmr_0x6068_nim3_io     [l_i] = 1;    // (TI, trig0_in) input on
        *p16_vmmr_0x606A_nim2_io     [l_i] = 1;    // (    trig1_in) input on
        *p16_vmmr_0x606E_nim0_io     [l_i] = 1;    // nim0 set to CBUS

        *p16_vmmr_0x6096_ts_sources  [l_i] = 0;    // clock from VME bus
        *p16_vmmr_0x6098_ts_divisor  [l_i] = 1;    // timestamp = time / ts_divisor

        // MMR (pulser test) settings
        *p16_vmmr_0x611A_bus_delay   [l_i] = 50;   // 50 = 10m of fiber bus length
        *p16_vmmr_0x6100_select_bus  [l_i] = 16;   // for all frontends

        *p16_vmmr_0x6110_fe_addr_wr  [l_i] = 0xC;  // address pulser register
        *p16_vmmr_0x6112_fe_data_wr  [l_i] = 0;    // switch local pulser off; Suggestion: use 0x6070 pulser reg.
        *p16_vmmr_0x6110_fe_addr_wr  [l_i] = 0x9;  // threshold register for lower 32 channels of one bank
        *p16_vmmr_0x6112_fe_data_wr  [l_i] = 35;   // (35 =about 10%)max value 255 = 70% full range 
        *p16_vmmr_0x6110_fe_addr_wr  [l_i] = 0xA;  // threshold register for higher 32 channels of one bank 
        *p16_vmmr_0x6112_fe_data_wr  [l_i] = 35;   // max value 255 = 70% full range 
        *p16_vmmr_0x6110_fe_addr_wr  [l_i] = 0xB;  // threshold register for channels 64-127 for mmr-128 
        *p16_vmmr_0x6112_fe_data_wr  [l_i] = 35;   // max value 255 = 70% full range 
        *p16_vmmr_0x6110_fe_addr_wr  [l_i] = 0x10; // data threshold register
        *p16_vmmr_0x6112_fe_data_wr  [l_i] = 200;  // = 5% full range; Directly works on ADC-range; Max value 4095
        //*p16_vmmr_0x6112_fe_data_wr  [l_i] = 10;  // = 5% full range; Directly works on ADC-range; Max value 4095

        // back to VMMR
        *p16_vmmr_0x6070_pulser_stat [l_i] = 1;    // 0 -> off, 1 -> on, 3 -> generate trigger but no ampl. data
        //*p16_vmmr_0x6070_pulser_stat [l_i] = 3;    // 0 -> off, 1 -> on, 3 -> generate trigger but no ampl. data
        *p16_vmmr_0x6072_pulser_ampl [l_i] = 0xFF; // maximum amplitude: 0xFF = 255
        *p16_vmmr_0x603A_start_acq   [l_i] = 0;    // stop acq
        *p16_vmmr_0x6090_reset_ctr_ab[l_i] = 3;    // reset counters
        *p16_vmmr_0x603C_fifo_reset  [l_i] = 1;    // fifo reset
        *p16_vmmr_0x6034_read_reset  [l_i] = 1;    // allow new trigger (gate?)
        *p16_vmmr_0x603A_start_acq   [l_i] = 1;    // start acq

        // 4-Jan-2024: new vmmr trigger features 
        *p16_vmmr_0x631C_no_docu     [l_i] = 0x1;  // define nim trigger group 0 for bus 0 (bits 0b1)
        *p16_vmmr_0x6320_no_docu     [l_i] = 0x2;  // define nim trigger group 1 for bus 1 (bits 0b10)
        *p16_vmmr_0x606E_nim0_io     [l_i] = 5;    // connect trigger group 0 to nim 0 out                         
        *p16_vmmr_0x606C_nim1_io     [l_i] = 5;    // connect trigger grop 1 to nim 1 out

        
        printm ("initialized VMMR id: %d \n", l_i); 
      }

      //------------------------------------- MDPP ---------------------------------------

      // initalize MDPP 
      for (l_i=0; l_i<N_MDPP; l_i++)
      {
        *p16_mdpp_0x6008_soft_reset  [l_i] = 1; sleep (1);
        *p16_mdpp_0x6004_mod_id      [l_i] = l_i;
        *p16_mdpp_0x6036_mult_evt    [l_i] = 0;    // single event mode
        *p16_mdpp_0x6038_mark_typ    [l_i] = 0;    // provide event counter, not time stamp
        *p16_mdpp_0x6032_dat_len_form[l_i] = 2;    // 32 bit data 
        *p16_mdpp_0x6044_out_form    [l_i] = 0;    // no other choice
        *p16_mdpp_0x603C_fifo_reset  [l_i] = 1;    // initialize fifo

        *p16_mdpp_0x605C_first_hit   [l_i] = 0;
        *p16_mdpp_0x6050_win_start   [l_i] = 16064; // 16064: 500 ns before trigger signal
        *p16_mdpp_0x6054_win_width   [l_i] = 640;   //   640: 1 us total (500 ns before, 500 ns after trigger)

        *p16_mdpp_0x6058_trig_src    [l_i] = 1;    // nim TI input trigger
 
        // channel settings
        *p16_mdpp_0x6100_sel_cha_pair[l_i] = 8;     // 8: select all channels
        *p16_mdpp_0x6124_shaping_time[l_i] = 32;   // set shaping time
        *p16_mdpp_0x6100_sel_cha_pair[l_i] = 0;     // 0: select channels  0,  1
        *p16_mdpp_0x611C_thresh0     [l_i] = 0x8fff;// set thresh channel  0
        *p16_mdpp_0x6100_sel_cha_pair[l_i] = 6;     // 6: select channels  12, 13
        *p16_mdpp_0x611C_thresh0     [l_i] = 0x2fff;// set thresh channel  12
        *p16_mdpp_0x6100_sel_cha_pair[l_i] = 7;     // 7: select channels  14, 15
        *p16_mdpp_0x611C_thresh0     [l_i] = 0xfff;  // set thresh channel 14
        *p16_mdpp_0x611E_thresh1     [l_i] = 0xfff;  // set thresh channel 15
        //

        *p16_mdpp_0x603A_start_acq   [l_i] = 0;    // stop acq
        *p16_mdpp_0x6090_reset_ctr_ab[l_i] = 3;    // reset counters
        *p16_mdpp_0x603C_fifo_reset  [l_i] = 1;    // fifo reset
        *p16_mdpp_0x6034_read_reset  [l_i] = 1;    // allow new trigger (gate?)
        *p16_mdpp_0x603A_start_acq   [l_i] = 1;    // start acq 

        printm ("initialized MDPP id: %d \n", l_i); 
      }
      
      //------------------------------------- V775 ---------------------------------------

      for (l_i=0; l_i<N_V775; l_i++)
      {
        // soft reset
        *p16_v775_0x1006_bit_set_1        [l_i] = 0x80;
        *p16_v775_0x1008_bit_clear_1      [l_i] = 0x80;
        *p16_v775_0x1008_bit_clear_1      [l_i] = 0x8;
        *p16_v775_0x1008_bit_clear_1      [l_i] = 0x10; 
        // standard setup
        *p16_v775_0x1002_geo_addr         [l_i] = 0x7;

        *p16_v775_0x1032_bit_set_2        [l_i] = 0x4;
        *p16_v775_0x1034_bit_clear_2      [l_i] = 0x4;
        *p16_v775_0x1034_bit_clear_2      [l_i] = 0x40;
        *p16_v775_0x1040_evt_ctr_reset    [l_i] = 0x1;

        *p16_v775_0x1034_bit_clear_2      [l_i] = 0x0;
        *p16_v775_0x1034_bit_clear_2      [l_i] = 0x1;
        *p16_v775_0x1034_bit_clear_2      [l_i] = 0x8;
        *p16_v775_0x1034_bit_clear_2      [l_i] = 0x10;
        *p16_v775_0x1034_bit_clear_2      [l_i] = 0x20;
        *p16_v775_0x1034_bit_clear_2      [l_i] = 0x40;
        *p16_v775_0x1034_bit_clear_2      [l_i] = 0x400;
        *p16_v775_0x1034_bit_clear_2      [l_i] = 0x1000;
        *p16_v775_0x1034_bit_clear_2      [l_i] = 0x2000;

        *p16_v775_0x1032_bit_set_2        [l_i] = 0x80;
        *p16_v775_0x1032_bit_set_2        [l_i] = 0x100;
        *p16_v775_0x1032_bit_set_2        [l_i] = 0x800;
        *p16_v775_0x1032_bit_set_2        [l_i] = 0x1000;
        *p16_v775_0x1032_bit_set_2        [l_i] = 0x4000;
 
        *p16_v775_0x1060_full_scale_range [l_i] = 0xff;
        *p16_v775_0x1060_full_scale_range [l_i] = 0x1e;
        // threshholds
        for (l_j=0; l_j<N_V775_CHA; l_j++)
        {
          *p16_v775_threshholds [l_i][l_j] = 0x0;
        }

        *p16_v775_0x1010_control          [l_i] = 0x0;

        printm ("initialized V775 id: %d \n", l_i); 
      }
      
      //------------------------------------- V785 ---------------------------------------

      for (l_i=0; l_i<N_V785; l_i++)
      {

        // soft reset
        *p16_v785_0x1006_bit_set_1        [l_i] = 0x80; 
        *p16_v785_0x1008_bit_clear_1      [l_i] = 0x80;
        *p16_v785_0x1008_bit_clear_1      [l_i] = 0x8;
        *p16_v785_0x1008_bit_clear_1      [l_i] = 0x10;
        // standard setup 
        *p16_v785_0x1002_geo_addr         [l_i] = 0x8;

        *p16_v785_0x1032_bit_set_2        [l_i] = 0x4;
        *p16_v785_0x1034_bit_clear_2      [l_i] = 0x4;
        *p16_v785_0x1034_bit_clear_2      [l_i] = 0x40;
        *p16_v785_0x1040_evt_ctr_reset    [l_i] = 0x1;

        *p16_v785_0x1034_bit_clear_2      [l_i] = 0x0;
        *p16_v785_0x1034_bit_clear_2      [l_i] = 0x1;
        *p16_v785_0x1034_bit_clear_2      [l_i] = 0x8;
        *p16_v785_0x1034_bit_clear_2      [l_i] = 0x10;
        *p16_v785_0x1034_bit_clear_2      [l_i] = 0x20;
        *p16_v785_0x1034_bit_clear_2      [l_i] = 0x40;
        *p16_v785_0x1034_bit_clear_2      [l_i] = 0x1000;
        *p16_v785_0x1034_bit_clear_2      [l_i] = 0x2000;

        *p16_v785_0x1032_bit_set_2        [l_i] = 0x80;
        *p16_v785_0x1032_bit_set_2        [l_i] = 0x100;
        *p16_v785_0x1032_bit_set_2        [l_i] = 0x800;
        *p16_v785_0x1032_bit_set_2        [l_i] = 0x1000;
        *p16_v785_0x1032_bit_set_2        [l_i] = 0x4000;
 
        //thresholds
        for (l_j=0; l_j<N_V785_CHA; l_j++)
        {
          *p16_v785_threshholds [l_i][l_j] = 70;
        }

        *p16_v785_0x1010_control          [l_i] = 0x0;

        printm ("initialized V785 id: %d \n", l_i); 
      }
      
      //------------------------------------- V830 ---------------------------------------

      for (l_i=0; l_i<N_V830; l_i++)
      {
        *p16_v830_0x1120_soft_reset[l_i] = 1;           // soft reset 
        *p16_v830_0x1108_control   [l_i] = 0x25;        // trigger - and header enable, 26 bit format 
        *p32_v830_0x1100_ena_chan  [l_i] = 0xffffffff;  // enable all channels
        *p16_v830_0x1110_geo_addr  [l_i] = 9;

        printm ("initialized V830 id: %d \n", l_i); 
      }

      //------------------------------------- MADC ---------------------------------------

      for (l_i=0; l_i<N_MADC; l_i++)
      {
        *p16_madc_0x6008_soft_reset  [l_i] = 1; usleep (500000);
//printm ("0x6004: %p\n", p16_madc_0x6008_soft_reset  [l_i]);        
        *p16_madc_0x6004_mod_id      [l_i] = l_i;
        *p16_madc_0x6036_mult_evt    [l_i] = 0;   // single event mode
        *p16_madc_0x6038_mark_typ    [l_i] = 0;   // provide event counter, not time stamp
        *p16_madc_0x6042_adc_resol   [l_i] = 4;   // 4: 8K, highes resolution, cvt: 12.5 us
        *p16_madc_0x6032_dat_len_form[l_i] = 2;   // 32 bit data 
        *p16_madc_0x6044_out_form    [l_i] = 0;   // no other choice
        *p16_madc_0x603C_fifo_reset  [l_i] = 1;   // initialize fifo
        *p16_madc_0x6090_reset_ctr_ab[l_i] = 3;   // reset counters

        for (l_j=0; l_j<N_MADC_CHA; l_j++)    // set thresholds
        {
//printm ("thresh: %p \n", p16_madc_thresholds[l_i][l_j]);        
          *p16_madc_thresholds[l_i][l_j] = 0x11;
        }

for (l_j=0; l_j<N_MADC_CHA; l_j++)    // set thresholds
{
  //printm ("thresh: 0x%x \n", *p16_madc_thresholds[l_i][l_j]);                  
}
        
        *p16_madc_0x6034_read_reset  [l_i] = 1;   // allow new trigger (gate?)
        *p16_madc_0x603A_start_acq   [l_i] = 1;   // accept gates from now on

        printm ("initialized MADC id: %d \n", l_i);
      }

      //------------------------------------- MTDC ---------------------------------------

      for (l_i=0; l_i<N_MTDC; l_i++)
      {
        *p16_mtdc_0x6008_soft_reset     [l_i] = 1; usleep (500000);
        *p16_mtdc_0x6004_mod_id         [l_i] = l_i;
        *p16_mtdc_0x6036_mult_evt       [l_i] = 0;   // single event mode
        *p16_mtdc_0x6038_mark_typ       [l_i] = 0;   // provide event counter, not time stamp
        *p16_mtdc_0x6042_tdc_resol      [l_i] = 4;   // 4: 15.6 ps
        *p16_mtdc_0x6032_dat_len_form   [l_i] = 2;   // 32 bit data 
        *p16_mtdc_0x6044_out_form       [l_i] = 0;
        *p16_mtdc_0x603C_fifo_reset     [l_i] = 1;   // initialize fifo
        *p16_mtdc_0x6068_trig_select    [l_i] = 0;   // NIM trigger 0 input       
        *p16_mtdc_0x6090_reset_ctr_ab   [l_i] = 3;   // reset counters

        *p16_mtdc_0x6034_read_reset     [l_i] = 1;   // allow new trigger (gate?)
        *p16_mtdc_0x603A_start_acq      [l_i] = 1;   // accept gates from now on

        printm ("initialized MTDC id: %d \n", l_i);
      }

      //------------------------------------- VUL_SCAL -----------------------------------

      for (l_i=0; l_i<N_VUL_SCAL; l_i++)
      {
        *p32_vul_scal_0x10_ena_cha [l_i] = 1;
        *p32_vul_scal_0x08_reset   [l_i] = 1;
        *p32_vul_scal_0x04_latch   [l_i] = 0;  // set to 0, regardless if hard - or soft latch is used
        printm ("initialized VULOM SCALER id: %d \n", l_i);     
      } 
      
      //------------------------------------- SIS3820 -----------------------------------

      for (l_i=0; l_i<N_SIS3820; l_i++)
      {
        *p32_sis3820_0x400_key_reset            [l_i] = 0x1;
        *p32_sis3820_0x404_key_sdram_fifo_reset [l_i] = 0x1;
        *p32_sis3820_0x100_operation_mode       [l_i] = 0x00220011; // [30:28] mode of operation 0 : "LATCHING SCALER MODE" mode : counter data is copied to shadow registers
                                                                    // [21:20] output mode       2 : ouptut 5 -> LNE pulse , outputs 6+7 -> 10 MHz pulses
                                                                    // [18:16] input mode        2 : input 1 -> external next pulse, input 3 -> inhibit counting 
                                                                    // [13:12] select SDRAM mode 1 : SDRAM (instead of FIFO) allows to read data from address 0x800000 
                                                                    // [ 6: 4] LNE (load next event) is taken from input 1 on front panel 
                                                                    // [    0] select non clearing mode 0: counters are NOT cleared at LNE        
        *p32_sis3820_0x418_key_operation_enable [l_i] = 0x1; // writing this register enables scaler operation
        printm ("initialized SIS3820 SCALER id: %d \n", l_i);     
      } 
      
      //----------------------------------- TRLOII ---------------------------------------
      #if defined USE_TRLOII
      trlo_init(&trlo);
      #endif

      //----------------------------------------------------------------------------------

      break;

    default:
      break;
  }

  fflush (stdout);

  return (1);
}

/*****************************************************************************/

int f_user_readout (CHARU    bh_trig_typ,
                    CHARU    bh_crate_nr,
                    INTS4   *pl_loc_hwacc,
                    INTS4   *pl_rem_cam,
                    INTS4   *pl_dat,
                    s_veshe *ps_veshe,
                    INTS4   *l_se_read_len,
                    INTS4   *l_read_stat)
{
  int   l_i, l_j; 
  short l_dat1, l_dat2;
  INTU4  l_dat;

  l_tr_ct[0]++;            // event/trigger counter
  l_tr_ct[bh_trig_typ]++;  // individual trigger counter
  
  pl_dat_save = (INTU4*)pl_dat; 
  *l_se_read_len = 0;

  switch (bh_crate_nr)
  {
    case 0 :
      switch (bh_trig_typ)
      {
        case 1:
        case 2:
        case 3:  
        case 4:       
        case 12:
        case 13:

          //------------------------------------- VETAR --------------------------------------

          #ifdef WR_TIME_STAMP
          if (l_check_wr_err == 1)
          {
            printm ("reset TLU fifo of white rabbit time stamp module vetar \n");
            f_wr_reset_tlu_fifo ();
            l_wr_init_ct++; 
            *l_read_stat = 0;               
            sleep (1);
            goto bad_event;
          }
          #endif //WR_TIME_STAMP

          #ifdef WR_TIME_STAMP  
          if (bh_trig_typ < 14)
          {
            *pl_dat++ = SUB_SYSTEM_ID;  // *l_se_read_len =+ 4;
    
            #ifdef WR_USE_TLU_DIRECT
            eb_stat_before = *fifo_ready;
            eb_fifo_ct_brd = *fifo_cnt; 
            *fifo_pop =      0xF;
            #if defined (RIO4) && defined (OSV_2_6_33)
            SERIALIZE_IO;
            #endif
            #ifdef IFC
            nanosleep(0);
            #endif 
            eb_tlu_high_ts = *ft_shi;
            eb_tlu_low_ts  = *ft_slo;
            eb_stat_after  = *fifo_ready;

            #else // WR_USE_TLU_DIRECT
    
            if ((eb_stat = eb_cycle_open(eb_device, 0, eb_block, &eb_cycle)) != EB_OK)
            {
              printm (RON"ERROR>>"RES" etherbone EP eb_cycle_open, status: %s \n", eb_status(eb_stat));
              //l_err_wr_ct++;
              //l_check_wr_err = 2; goto bad_event; 
            }

            /* Queueing operations to a cycle never fails (EB_OOM is reported later) */
            /* read status of FIFOs */
            eb_cycle_read (eb_cycle, wrTLU + GSI_TM_LATCH_FIFO_READY, EB_BIG_ENDIAN|EB_DATA32, &eb_stat_before);
            /* read fifo fill counter */
            eb_cycle_read (eb_cycle, wrTLU + GSI_TM_LATCH_FIFO_CNT, EB_BIG_ENDIAN|EB_DATA32, &eb_fifo_ct_brd);
            /* pop timestamp from FIFO */
            eb_cycle_write (eb_cycle, wrTLU + GSI_TM_LATCH_FIFO_POP, EB_BIG_ENDIAN|EB_DATA32, 0xF);
            /* read high word of latched timestamp */
            eb_cycle_read (eb_cycle, wrTLU + GSI_TM_LATCH_FIFO_FTSHI, EB_BIG_ENDIAN|EB_DATA32, &eb_tlu_high_ts);
            /* read low word of latched timestamp */
            eb_cycle_read (eb_cycle, wrTLU + GSI_TM_LATCH_FIFO_FTSLO, EB_BIG_ENDIAN|EB_DATA32, &eb_tlu_low_ts);

            /* read status of FIFOs */
            eb_cycle_read (eb_cycle, wrTLU + GSI_TM_LATCH_FIFO_READY, EB_BIG_ENDIAN|EB_DATA32, &eb_stat_after);

            if ((eb_stat = eb_cycle_close(eb_cycle)) != EB_OK)
            {
              printm (RON"ERROR>>"RES" etherbone EP eb_cycle_close, status: %s \n", eb_status(eb_stat));
              //l_err_wr_ct++;
              //l_check_wr_err = 2; goto bad_event; 
            }
            #endif //WR_USE_TLU_DIRECT

            if ((eb_stat_before & l_used_tlu_fifo) != l_used_tlu_fifo)
            {
              usleep (100000);
              printm (RON"ERROR>>"RES" TLU fifo %d is empty before time stamp read, stat: 0x%x\n", WR_TLU_FIFO_NR, eb_stat_before);
              l_err_wr_ct++;
              l_check_wr_err = 2; goto bad_event; 
            }
            if ((eb_stat_after & l_used_tlu_fifo) != 0)
            {
              //printm (RON"ERROR>>"RES" TLU fifo %d is not empty after time stamp read, stat: 0x%x\n", WR_TLU_FIFO_NR, eb_stat_after);
            }

            #ifdef USER_TRIG_CLEAR
            if (eb_fifo_ct_brd > 2)
            {
              printm (RON"ERROR>>"RES" TLU fill count: %d is bigger than 2 \n", eb_fifo_ct_brd);
              l_err_wr_ct++;
              l_check_wr_err = 2; goto bad_event; 
            }
            #else
            if (eb_fifo_ct_brd > 1)
            {
              printm (RON"ERROR>>"RES" TLU fill count: %d is bigger than 1 \n", eb_fifo_ct_brd);
              l_err_wr_ct++;
              l_check_wr_err = 2; goto bad_event; 
            }
            #endif // USER_TRIG_CLEAR

            ll_ts_hi = (unsigned long long) (eb_tlu_high_ts & 0xFFFFFFFF);
            ll_ts_lo = (unsigned long long) (eb_tlu_low_ts  & 0xFFFFFFFF);
     
            ll_ts_hi = ll_ts_hi << 35;
            ll_ts_lo = ll_ts_lo <<  3;
            ll_ts = 0;
            ll_ts        = ll_ts_hi + ll_ts_lo;
            ll_timestamp = ll_ts;

            ll_l16   = (ll_ts >>  0) & 0xffff;
            ll_m16   = (ll_ts >> 16) & 0xffff; 
            ll_h16   = (ll_ts >> 32) & 0xffff;
            ll_x16   = (ll_ts >> 48) & 0xffff;   

            l_time_l16 = (unsigned long) (ll_l16);
            l_time_m16 = (unsigned long) (ll_m16);
            l_time_h16 = (unsigned long) (ll_h16);
            l_time_x16 = (unsigned long) (ll_x16);

            l_time_l16 = (l_time_l16 & 0xffff) + (TS__ID_L16 << 16);
            l_time_m16 = (l_time_m16 & 0xffff) + (TS__ID_M16 << 16);
            l_time_h16 = (l_time_h16 & 0xffff) + (TS__ID_H16 << 16);
            l_time_x16 = (l_time_x16 & 0xffff) + (TS__ID_X16 << 16);

            *pl_dat++ = l_time_l16;
            *pl_dat++ = l_time_m16;
            *pl_dat++ = l_time_h16;
            *pl_dat++ = l_time_x16;

            ll_actu_timestamp = ll_timestamp;
            ll_diff_timestamp = ll_actu_timestamp - ll_prev_timestamp;   
            if (ll_prev_timestamp > ll_actu_timestamp)
            {
              printm ("ERROR>> actual time stamp earlier than previous one \n");
              printm ("        actual   ts: 0x%llx \n", ll_actu_timestamp);
            }

            ll_prev_timestamp     = ll_actu_timestamp;
            eb_tlu_high_ts_prev   = eb_tlu_high_ts;
            eb_tlu_low_ts_prev    = eb_tlu_low_ts;
          } 
          #endif // WR_TIME_STAMP  

          if ( (l_tr_ct[0] % STATISTIC) == 0)
          {
            printm ("----------------------------------------------------\n");
            printm ("nr of triggers processed: %u \n", l_tr_ct[0]);
            printm ("\n");
            for (l_i=1; l_i<MAX_TRIG_TYPE; l_i++)
            {
              if (l_tr_ct[l_i] != 0)
              {
                printm ("trigger type %2u found %10u times \n", l_i, l_tr_ct[l_i]);
              }
            }
            #ifdef WR_TIME_STAMP
            printm ("");
            printm ("reset White Rabbit VETAR TLU fifo      %d times \n", l_wr_init_ct);
            #endif // WR_TIME_STAMP 
            printm ("----------------------------------------------------\n");  
          }
          
          bad_event:

          #ifdef WR_TIME_STAMP  
          if (l_check_wr_err != 0)
          {
            if (l_check_wr_err == 2)
            {
              printm ("white rabbit failure: invalidate current trigger/event (1) (0xbad00bad)\n");
            }
            if (l_check_wr_err == 1)
            {
              printm ("white rabbit failure: invalidate current trigger/event (2) (0xbad00bad)\n");
              printm ("");
            }
            pl_dat    = (INTU4*)pl_dat_save;
            *pl_dat++ = 0xbad00bad;
            l_check_wr_err--;
          }
          #endif // WR_TIME_STAMP

          //------------------------------------- VMMR ---------------------------------------         

          for (l_i=0; l_i<N_VMMR; l_i++)
          {
            l_dat1 = *p16_vmmr_0x6004_mod_id[l_i];
            l_dat2 = *p16_vmmr_0x600E_firm_rev[l_i];
            l_dat2 = 0; // problem since 11-feb-2022 with firm rev
                        // new vmmr ? not likely || r4l-14 instead of r4l-15 ??
            #ifdef DEBUG_SLOW
            printm ("mod id:   0x%x \n", l_dat1);
            printm ("firm rev: 0x%x \n", l_dat2);
            #endif
            l_dat  =  (INTU4)l_dat2 + (((INTU4)l_dat1) << 16) | 0xA0000000; // 0xA0000000 = VMMR
	    //printm ("modid, firmware revision: 0x%x  \n", l_dat); sleep (1);
            *pl_dat++ = l_dat;


            // check for data ready
            l_ct = 0; 
            l_dat1 = *p16_vmmr_0x603E_dat_ready[l_i];
            //printm ("VMMR data ready: 0x%x \n", l_dat1);
            l_dat1 = l_dat1 & 0x1;
            while (l_dat1 != 1)
            {
              l_dat1 = *p16_vmmr_0x603E_dat_ready[l_i];
              l_ct++;
              sched_yield ();
              if (l_ct > 10)
              {
                sleep(1);
                printm ("data not yet ready from VMMR id: %d, register value was 0x%x \n", l_i, l_dat1);
              }
              l_dat1 = *p16_vmmr_0x603E_dat_ready[l_i];
            }

            // read data length
            l_dat1 = *p16_vmmr_0x6030_buf_dat_len[l_i];
            // readout VMMR
            for (l_j=0; l_j<l_dat1; l_j++)
            {
              *pl_dat++ = *pl_vmmr_dat[l_i];
            }
            // allow for new trigger
            *p16_vmmr_0x6034_read_reset[l_i] = 1;
          }

          //------------------------------------- MDPP ---------------------------------------

          for (l_i=0; l_i<N_MDPP; l_i++)
          {
            l_dat1 = *p16_mdpp_0x6004_mod_id[l_i];
            l_dat2 = *p16_mdpp_0x600E_firm_rev[l_i];
            #ifdef DEBUG_SLOW
            printm ("mod id:   0x%x \n", l_dat1);
            printm ("firm rev: 0x%x \n", l_dat2);
            #endif
            l_dat  =  (INTU4)l_dat2 + (((INTU4)l_dat1) << 16) | 0xB0000000; // 0xB0000000 = MDPP
	    //printm ("modid, firmware revision: 0x%x  \n", l_dat); sleep (1);
            *pl_dat++ = l_dat;

            // check for data ready
            l_ct = 0; 

            l_dat1 = *p16_mdpp_0x603E_dat_ready[l_i];
            #ifdef DEBUG_SLOW
            printm ("First data ready is :   0x%x \n", l_dat1);
            #endif
            while (l_dat1 != 1)
            {
              l_dat1 = *p16_mdpp_0x603E_dat_ready[l_i];
              l_ct++;
              sched_yield ();
              if (l_ct > 10)
              {
                sleep(1);
                printm ("data not yet ready from MDPP id: %d, register value was 0x%x \n", l_i, l_dat1);
              }
              l_dat1=*p16_mdpp_0x603E_dat_ready[l_i];
            }

            // read data length
            l_dat1 = *p16_mdpp_0x6030_buf_dat_len[l_i];
            #ifdef DEBUG_SLOW
            printm ("data size: 0x%x \n", l_dat1);
            #endif

            // readout MDPP
            for (l_j=0; l_j<l_dat1; l_j++)
            {
              *pl_dat++ = *pl_mdpp_dat[l_i]; 
            }
            // allow for new trigger
            *p16_mdpp_0x6034_read_reset[l_i] = 1;
          }

          //------------------------------------- V775 ---------------------------------------

          for (l_i=0; l_i<N_V775; l_i++)
          {
            l_dat1 = *p16_v775_0x1002_geo_addr [l_i] & 0x1f;
            l_dat2 = *p16_v775_0x1000_firm_rev [l_i];
            l_dat  =  (INTU4)l_dat2 + (((INTU4)l_dat1) << 16) | 0xC0000000; // 0xC0000000 = V775
	    //printm ("modid, firmware revision: 0x%x  \n", l_dat); sleep (1);
            *pl_dat++ = l_dat;

            l_dat1 = (*p16_v775_0x100e_status_1 [l_i]) & 0x1;
            //printm ("status 1: 0x%x \n", l_dat1);
            if (l_dat1 != 1)
            {
              printm ("ERROR>> data from v775, id: %d not ready, exiting.. \n", l_i);
              exit (1);
            }
        
            l_v775_head = (INTU4)*pl_v775_dat[l_i];
            //printm ("l_v775_head: 0x%x \n", l_v775_head); sleep (1);
            *pl_dat++ = l_v775_head;
         
            l_dat = ((l_v775_head >> 8) & 0x3f) + 1; // data size
            for (l_j=0; l_j<l_dat; l_j++)
            {
              *pl_dat++ = *pl_v775_dat[l_i];
            }

            l_dat1 = (*p16_v775_0x100e_status_1 [l_i]) & 0x1;
            if (l_dat1 == 1)
            {
              printm ("ERROR>> data from v775, id: %d module not empty; resetting.. \n", l_i);
              sleep (1);
              *p16_v775_0x1032_bit_set_2   [l_i] = 0x4;
              *p16_v775_0x1034_bit_clear_2 [l_i] = 0x4;
            }
          }

          //------------------------------------- V785 ---------------------------------------

          for (l_i=0; l_i<N_V785; l_i++)
          {
            l_dat1 = *p16_v785_0x1002_geo_addr [l_i] & 0x1f;
            l_dat2 = *p16_v785_0x1000_firm_rev [l_i];
            l_dat  =  (INTU4)l_dat2 + (((INTU4)l_dat1) << 16) | 0xD0000000; // 0xD0000000 = V785
	    //printm ("modid, firmware revision: 0x%x  \n", l_dat); sleep (1);
            *pl_dat++ = l_dat;

            l_dat1 = (*p16_v785_0x100e_status_1 [l_i]) & 0x1;
            //printm ("status 1: 0x%x \n", l_dat1);
            if (l_dat1 != 1)
            {
              printm ("ERROR>> data from v785, id: %d not ready, exiting.. \n", l_i);
              exit (1);
            }
        
            l_v785_head = (INTU4)*pl_v785_dat[l_i];
            //printm ("l_v785_head: 0x%x \n", l_v785_head); sleep (1);
            *pl_dat++ = l_v785_head;
         
            l_dat = ((l_v785_head >> 8) & 0x3f) + 1; // data size
            for (l_j=0; l_j<l_dat; l_j++)
            {
              *pl_dat++ = *pl_v785_dat[l_i];
            }

            l_dat1 = (*p16_v785_0x100e_status_1 [l_i]) & 0x1;
            if (l_dat1 == 1)
            {
              printm ("ERROR>> data from v785, id: %d module not empty; resetting.. \n", l_i);
              sleep (1);
              *p16_v785_0x1032_bit_set_2   [l_i] = 0x4;
              *p16_v785_0x1034_bit_clear_2 [l_i] = 0x4;
            }
          }

          //------------------------------------- V830 ---------------------------------------

          for (l_i=0; l_i<N_V830; l_i++)
          {
            l_dat1 = *p16_v830_0x1110_geo_addr [l_i] & 0x1f;
            l_dat2 = *p16_v830_0x1132_firm_rev [l_i];
            l_dat  =  (INTU4)l_dat2 + (((INTU4)l_dat1) << 16) | 0xE0000000; // 0xE0000000 = V830
	    //printm ("modid, firmware revision: 0x%x  \n", l_dat); sleep (1);
            *pl_dat++ = l_dat;

            l_dat1 = (*p16_v830_0x110E_status [l_i]) & 0x1;
            //printm ("status 1: 0x%x \n", l_dat1);
            if (l_dat1 != 1)
            {
              printm ("ERROR>> data from v830, id: %d not ready, exiting.. \n", l_i);
              exit (1);
            }

            for (l_j=0; l_j<(N_V830_CHA+1); l_j++)  // there is no trailer
            {
              *pl_dat++ = *pl_v830_dat[l_i];
            }

            l_dat1 = (*p16_v830_0x110E_status [l_i]) & 0x1;
            if (l_dat1 == 1)
            {
              printm ("ERROR>> data from v830, id: %d module not empty \n", l_i);
              sleep (1);
            }
          }

          //------------------------------------- MADC ---------------------------------------

          for (l_i=0; l_i<N_MADC; l_i++)
          {
            l_dat1 = *p16_madc_0x6004_mod_id[l_i];
            l_dat2 = *p16_madc_0x600E_firm_rev[l_i];
            #ifdef DEBUG_SLOW
            printm ("mod id:   0x%x \n", l_dat1);
            printm ("firm rev: 0x%x \n", l_dat2);
            #endif
            l_dat  =  (INTU4)l_dat2 + (((INTU4)l_dat1) << 16) | 0xA1000000; // 0xA1000000 = MADC
	    //printm ("modid, firmware revision: 0x%x  \n", l_dat); sleep (1);
            *pl_dat++ = l_dat;

            // check for data ready
            l_ct = 0; 

            l_dat1 = *p16_madc_0x603E_dat_ready[l_i];
            #ifdef DEBUG_SLOW
            printm ("First data ready is :   0x%x \n", l_dat1);
            #endif
            while (l_dat1 != 1)
            {
              l_dat1 = *p16_madc_0x603E_dat_ready[l_i];
              l_ct++;
              sched_yield ();
              if (l_ct > 10)
              {
                sleep(1);
                printm ("data not yet ready from MADC id: %d, register value was 0xd \n", l_i, l_dat1);
              }
              l_dat1=*p16_madc_0x603E_dat_ready[l_i];
            }

            // read data length
            l_dat1 = *p16_madc_0x6030_buf_dat_len[l_i];
            #ifdef DEBUG_SLOW
            printm ("data size: 0x%x \n", l_dat1);
            #endif

            // readout MADC
            for (l_j=0; l_j<l_dat1; l_j++)
            {
              *pl_dat++ = *pl_madc_dat[l_i]; 
            }
            // allow for new trigger
            *p16_madc_0x6034_read_reset[l_i] = 1;
          }

          //------------------------------------- MTDC ---------------------------------------

          for (l_i=0; l_i<N_MTDC; l_i++)
          {
            l_dat1 = *p16_mtdc_0x6004_mod_id[l_i];
            l_dat2 = *p16_mtdc_0x600E_firm_rev[l_i];
            #ifdef DEBUG_SLOW
            printm ("mod id:   0x%x \n", l_dat1);
            printm ("firm rev: 0x%x \n", l_dat2);
            #endif
            l_dat  =  (INTU4)l_dat2 + (((INTU4)l_dat1) << 16) | 0xB1000000; // 0xB1000000 = MTDC
	    //printm ("modid, firmware revision: 0x%x  \n", l_dat); sleep (1);
            *pl_dat++ = l_dat;

            // check for data ready
            l_ct = 0; 

            l_dat1 = *p16_mtdc_0x603E_dat_ready[l_i];

            #ifdef DEBUG_SLOW
            printm ("First data ready is :   0x%x \n", l_dat1);
            #endif
            while (l_dat1 != 1)
            {
              l_dat1 = *p16_mtdc_0x603E_dat_ready[l_i];
              l_ct++;
              sched_yield ();
              if (l_ct > 10000)
              {
                sleep(1);
                printm ("data not yet ready from MTDC id: %d, register value was 0xd \n", l_i, l_dat1);
              }
              l_dat1=*p16_mtdc_0x603E_dat_ready[l_i];
            }

            // read data length
            l_dat1 = *p16_mtdc_0x6030_buf_dat_len[l_i];
            #ifdef DEBUG_SLOW
            printm ("data size: 0x%x \n", l_dat1);
            #endif

            // readout MTDC
            for (l_j=0; l_j<l_dat1; l_j++)
            {
              *pl_dat++ = *pl_mtdc_dat[l_i]; 
            }
            // allow for new trigger
            *p16_mtdc_0x6034_read_reset[l_i] = 1;
          }

          //------------------------------------- VUL_SCAL -----------------------------------

          for (l_i=0; l_i<N_VUL_SCAL; l_i++)
          {
            *pl_dat++ = (l_i << 16) | 0xC1000000;
            
            // readout VUL_SCAL
            //*p32_vul_scal_0x04_latch [l_i] = 1;  // for software latch used. comment this line, if hardware latch is used.
            *pl_dat = *p32_vul_scal_dat[l_i][0]; // required because of latch issue.
            for (l_j=0; l_j<N_VUL_SCAL_CHA; l_j++)
            {
              //usleep (1); 
              //printm ("bloed %d %d %p \n", l_i, l_j, p32_vul_scal_dat[l_i][l_j]);
              *pl_dat++ = *p32_vul_scal_dat[l_i][l_j];
            }
          }

          //------------------------------------- SIS3820 -----------------------------------

          for (l_i=0; l_i<N_SIS3820; l_i++)
          {
            *pl_dat++ = (l_i << 16) | 0xD1000000 | l_sis3820_channels[l_i]; // SIS3820 magic word and number of scaler channelsS
            #if defined IFC 
            // The IFC shows strange behavior when accessing the counter shadow registers directly (the second read delivers garbage data).
            // This workaround fixes the issue with two dummy reads before the actual readout.
            *pl_dat   = *p32_sis3820_0x800_counter_shadow[l_i][0];
            *pl_dat   = *p32_sis3820_0x800_counter_shadow[l_i][0];
            #endif
            for (l_j=0; l_j<N_SIS3820_CHA; l_j++)
            {
              if (0x1&(SIS3820_MASK>>l_j)) *pl_dat++ = *p32_sis3820_0x800_counter_shadow[l_i][l_j];
            }
          }

          //----------------------------------- TRLOII ---------------------------------------

          #if defined USE_TRLOII
	  trlo_readout(&trlo, (uint32_t **)&pl_dat);
          #endif

          //----------------------------------------------------------------------------------
          break;
        case 15:
          #ifdef IFC
          // f_ifc_a32_vme_mas_free ((char*)p_TLU,                 &alt_win_vetar);
          // f_ifc_a32_vme_mas_free ((char*)p16_virt_vme_mod_base, &alt_win_modules);
          #endif // IFC
          break;
        default:
          break;
      }
      break;                      
    default:
      break;
  }

  *l_se_read_len = (INTS4)((long)pl_dat - (long)pl_dat_save);
  #ifdef DEBUG_SLOW
  sleep(1);
  #endif

  return (1);
}
/****************************************************************************/

#ifdef IPV
void f_ipv_a32_vme_mas_map ()
{
  pev = pev_init(1);
  if (!pev)
  {
    printm("ERROR>> cannot allocate data structures to control PEV1100\n");
    exit( -1);
  }
  if (pev->fd < 0)
  {
    printm("ERROR>> cannot find PEV1100 interface\n");
    exit( -1);
  }

  vme_mas_map.rem_addr = VME_BASE;
  vme_mas_map.mode  = MAP_ENABLE | MAP_ENABLE_WR | MAP_SPACE_VME | MAP_VME_A32;
  vme_mas_map.flag  = 0x0;
  vme_mas_map.sg_id = MAP_MASTER_32;
  vme_mas_map.size  = N_VMMR * VMMR_OFF;
  pev_map_alloc (&vme_mas_map);


  shm_vme_addr = pev_mmap( &vme_mas_map);
  if( shm_vme_addr == MAP_FAILED)
  {
    printm ("ERROR>> A32 VME mapping failed\n");
    
  }
  p16_vmmr_virt_vme_base = (short*) shm_vme_addr;
}

void f_ipv_a32_vme_mas_free ()
{
  pev_map_free (&vme_mas_map);
  pev_exit (pev);
}
#endif // IPV

/****************************************************************************/

#ifdef WR_TIME_STAMP
void f_wr_init ()  
{
  #ifdef WR_USE_TLU_DIRECT
  *ch_select = WR_TLU_FIFO_NR;  
  printm ("directly selected White Rabbit TLU FIFO channel number: %3d \n", *ch_select);
  eb_fifo_size = *ch_fifosize;
  printm ("size of  White Rabbit TLU FIFO:                         %3d \n", eb_fifo_size);
  *fifoclear =0xFFFFFFFF;
  *armset =0xFFFFFFFF;

  #else //  WR_USE_TLU_DIRECT 

  // select FIFO channel
  if ((eb_stat = eb_device_write (eb_device, wrTLU + GSI_TM_LATCH_CH_SELECT, EB_BIG_ENDIAN|EB_DATA32, WR_TLU_FIFO_NR, 0, eb_block)) != EB_OK)
  {
    printm (RON"ERROR>>"RES" when selecting TLU FIFO channel \n");
  }

  // read back selected FIFO channel
  if ((eb_stat = eb_device_read (eb_device, wrTLU + GSI_TM_LATCH_CH_SELECT, EB_BIG_ENDIAN|EB_DATA32, &eb_fifo_cha, 0, eb_block)) != EB_OK)
  {
    printm (RON"ERROR>>"RES" when reading selected FIFO channel number \n");
  }
  else
  {
    printm ("");
    printm ("selected White Rabbit TLU FIFO channel number: %3d \n", eb_fifo_cha);
  }

  l_eb_first2 = 1;
  // size of TLU FIFO
  if ((eb_stat = eb_device_read (eb_device, wrTLU + GSI_TM_LATCH_CHNS_FIFOSIZE, EB_BIG_ENDIAN|EB_DATA32, &eb_fifo_size, 0, eb_block)) != EB_OK)
  {
    printm (RON"ERROR>>"RES" when reading TLU FIFO size \n");
  }
  else
  {
    printm ("size of  White Rabbit TLU FIFO:                %3d \n", eb_fifo_size);
    printm ("");
  }

  /* prepare the TLU for latching of timestamps */
  /* clear all FIFOs */
  if ((eb_stat = eb_device_write(eb_device, wrTLU + GSI_TM_LATCH_FIFO_CLEAR,
                                          EB_BIG_ENDIAN|EB_DATA32, 0xFFFFFFFF, 0, eb_block)) != EB_OK)
  {
    printm (RON"ERROR>>"RES" etherbone TLU eb_device_write (CLEAR TLU FIFOS), status: %s \n", eb_status(eb_stat));
  }

  /* arm triggers for latching */
  if ((eb_stat = eb_device_write(eb_device, wrTLU + GSI_TM_LATCH_TRIG_ARMSET,
                                          EB_BIG_ENDIAN|EB_DATA32, 0xFFFFFFFF, 0, eb_block)) != EB_OK)
  {
    printm (RON"ERROR>>"RES" etherbone TLU eb_device_write (ARM LATCHING), status: %s \n", eb_status(eb_stat));
  }
  #endif  //  WR_USE_TLU_DIRECT
}
#endif // WR_TIME_STAMP

/*****************************************************************************/

#ifdef WR_TIME_STAMP
void f_wr_reset_tlu_fifo ()  
{
  /* clear all FIFOs */
  #ifdef WR_USE_TLU_DIRECT
  *fifoclear =0xFFFFFFFF;
  #else  
  if ((eb_stat = eb_device_write(eb_device, wrTLU + GSI_TM_LATCH_FIFO_CLEAR,
                                          EB_BIG_ENDIAN|EB_DATA32, 0xFFFFFFFF, 0, eb_block)) != EB_OK)
  {
    printm (RON"ERROR>>"RES" etherbone TLU eb_device_write (CLEAR TLU FIFOS), status: %s \n", eb_status(eb_stat));
  }
  #endif
}
#endif // WR_TIME_STAMP

/****************************************************************************/

#ifdef IFC
void f_ifc_a32_vme_mas_map (int l_vme_base, int l_vme_size, struct alt_ioctl_map_win *alt_win)
{
  long offset = 0, len = 0;
  int alt_fd = alt_init ();    // does not hurt if it has been called before :)
  if (alt_fd<=0)
  {
    printm ("f_map_ifc_a32_alt_vme - Error: could not do alt_init\n");
    exit (-1);
  }

  printm ("base: 0x%x, size: 0x%x \n", l_vme_base, l_vme_size);
  
  alt_win->req.rem_addr   = l_vme_base;
  alt_win->req.loc_addr   = MAP_LOC_ADDR_AUTO; //=-1;
  alt_win->req.size       = l_vme_size;
  alt_win->req.mode.sg_id = MAP_ID_MAS_PCIE_PMEM;
  alt_win->req.mode.space = MAP_SPACE_VME;
  alt_win->req.mode.am    = 0x9;
  alt_win->req.mode.swap  = 0; //MAP_SPLIT_D32; // MAP_SWAP_AUTO;
  alt_win->req.mode.flags = MAP_FLAG_FREE;
  if (alt_map_alloc (alt_win) < 0)
  {
    printm ("f_map_ifc_a32_mas_vme: ERROR>> alt_map_alloc failed, exiting..\n");
    exit (-1);
  }
  offset =  l_vme_base - alt_win->req.rem_addr;
  printm ("Got althea VME window parameters: vme_addr: 0x%lx loc_addr: 0x%lx size: %x, offset:0x%lx \n",
  alt_win->req.rem_addr, alt_win->req.loc_addr, alt_win->req.size , offset);
  // map local address to user space
  offset += alt_win->req.loc_addr; // take mapped base into acount
  len = alt_win->req.size;

  // here we use our own trigmod driver for mmap, because we have put the noncached option into it :)
  triva_fd = open (triva_driverhandle, O_RDWR);
  if (triva_fd < 0)
  {
    printm ("ERROR>> cannot open TRIGMOD device for mapping %s\n", triva_driverhandle);
    printm ("Error>> %s\n", strerror (errno));
    exit (-1);
  }

  p16_virt_vme_base = (short*) mmap (NULL, len, PROT_READ | PROT_WRITE, MAP_SHARED, triva_fd, offset);
  if (p16_virt_vme_base == MAP_FAILED)
  {
    printm ("f_map_ifc_a32_mas_vme ERROR>> mmap failed, cannot get user space address. exiting..\n");
    exit (-1);
  }
  printm ("Mapped to user space address p16_virt_vme_base = %p \n", p16_virt_vme_base);
}

void f_ifc_a32_vme_mas_free (char* p_virt_vme_base, struct alt_ioctl_map_win *alt_win)
{
  // JAM 5-2021: note that this function is never called. Should be something like f_user_exit() later?
  if (p_virt_vme_base > 0)
  munmap (p_virt_vme_base, alt_win->req.size);
  if (alt_map_free(alt_win)<0)
  {
    printm ("WARNING: alt_map_free failed.\n");
  }
  //close (alt_fd);
}
#endif // IFC

/****************************************************************************/

int f_user_exit ()
{
  printf ("My f_user_exit ()...");

  #ifdef IFC
   #ifdef WR_USE_TLU_DIRECT
    f_ifc_a32_vme_mas_free ((char*)p_TLU, &alt_win_vetar);
   #endif
   f_ifc_a32_vme_mas_free ((char*)p16_virt_vme_mod_base, &alt_win_modules);  
   #ifdef USE_TRLOII
   trlo_unmap_hardware(trlo._trlo_unmapinfo);
   #endif
  #endif // IFC

  #ifdef WR_USE_TLU_DIRECT_WITHOUT_KERNEL_MODULE  
   f_a32_vme_mas_free_TLU ();
   f_a24_vme_mas_free_CONTROL ();
   f_a32_vme_mas_free_CRCSR ();
  #endif // WR_USE_TLU_DIRECT_WITHOUT_KERNEL_MODULE

  printf ("the End!\n");
  return (1);
}

/***************************************************************************************************************/

#ifdef WR_USE_TLU_DIRECT_WITHOUT_KERNEL_MODULE

/***************************************************************************************************************/

void f_a32_vme_mas_map_TLU ()
{
  #ifdef RIO2
  // TODO: meaning of this parameters?
  param.iack  = 1;
  param.rdpref = 0;
  param.wrpost = 0;
  param.swap = SINGLE_AUTO_SWAP;
  vmeaddr = VETAR_REGS_ADDR;
  vme_len = VETAR_REGS_SIZE;
  vme_am = 0x09;
  pl_virt_vme_base_tlu  = (INTU4*) find_controller(vmeaddr,vme_len,vme_am,0,0,&param);
  if( pl_virt_vme_base_tlu ==  (INTU4*) -1 )
  {
    printm ("f_map_a32_mas_vme ERROR>> find_controller() failed, cannot get user space address. exiting...\n");
    exit (-1);
  }
  #endif

  printm ("Mapped to user space address pl_virt_vme_base_tlu= %p \n", pl_virt_vme_base_tlu);
}

/***************************************************************************************************************/

void f_a32_vme_mas_free_TLU ()
{
  #ifdef RIO2
  return_controller(pl_virt_vme_base_tlu,VETAR_REGS_SIZE);
  #endif
}

/*****************************************************************************/
void f_a24_vme_mas_map_CONTROL ()
{
  #ifdef RIO2
  // TODO: meaning of this parameters?
  param.iack  = 1;
  param.rdpref = 0;
  param.wrpost = 0;
  param.swap = SINGLE_AUTO_SWAP;
  vmeaddr = VETAR_CTRL_ADDR;
  vme_len = VETAR_CTRL_SIZE;
  vme_am = 0x39;//VME_AM_A24 | VME_AM_DATA;
  pl_virt_vme_base_control  = (INTU4*) find_controller(vmeaddr,vme_len,vme_am,0,0,&param);
  if( pl_virt_vme_base_control ==  (INTU4*) -1 )
  {
    printm ("f_a24_vme_mas_map_CONTROL ERROR>> find_controller() failed, cannot get user space address. exiting...\n");
    exit (-1);
  }
  #endif
  
  printm ("Mapped to user space address pl_virt_vme_base_control= %p \n", pl_virt_vme_base_control);

}

/***************************************************************************************************************/

void f_a24_vme_mas_free_CONTROL ()
{
  #ifdef RIO2
  return_controller(pl_virt_vme_base_control,VETAR_CTRL_SIZE);
  #endif
}

/*****************************************************************************/
void f_a32_vme_mas_map_CRCSR ()
{
  #ifdef RIO2
  param.iack  = 1;
  param.rdpref = 0;
  param.wrpost = 0;
  param.swap = SINGLE_AUTO_SWAP;
  vmeaddr = VETAR_CRCSR_ADDR;
  vme_len = VETAR_CRCSR_SIZE;
  vme_am = 0x2F;    //VME_AM_CRCSR;
  // JAM RIO3: extend test already on mapping time:
  offset=VME_VENDOR_ID_OFFSET; //0
  size=4; // 0
  pl_virt_vme_base_crcsr  = (INTU4*) find_controller(vmeaddr,vme_len,vme_am,offset,size,&param);
  if( pl_virt_vme_base_crcsr ==  (INTU4*) -1 )
  {
    printm ("f_a32_vme_mas_map_CRCSR ERROR>> find_controller() failed, cannot get user space address. exiting...\n");
    //pl_virt_vme_base_crcsr=0;
    exit (-1);
  }
  printm ("f_a32_vme_mas_map_CRCSR has mapped VME address 0x%x to:0x%x, param.sgmin=0x%x, testing with off:0x%x, size=%d",
                                                              vmeaddr, pl_virt_vme_base_crcsr, param.sgmin, offset, size);
  #endif

  printm ("Mapped to user space address pl_virt_vme_base_crcsr= %p \n", pl_virt_vme_base_crcsr);
}

/***************************************************************************************************************/

void f_a32_vme_mas_free_CRCSR ()
{
  #ifdef  RIO2
  return_controller(pl_virt_vme_base_crcsr,VETAR_CRCSR_SIZE);
  #endif
}

/***************************************************************************************************************/

INTS4* f_set_csr_address (int off)
{
  INTS4* address = 0;
  off -= off % 4;    // JAM22: account address boundary of crcsr
  address = (INTS4*) ((char*) (pl_virt_vme_base_crcsr) + off);
  return address;
}

/***************************************************************************************************************/

int f_vetar_is_present ()
{
  int idc;
  printm ("Check if VETAR is present at config base address 0x%x\n", VETAR_CRCSR_ADDR);
  idc = (*cr_vendor_id << 16);
  idc += (*(cr_vendor_id + 1) << 8);
  idc += (*(cr_vendor_id + 2));
  if (idc == VETAR_VENDOR_ID)
  {
    printm ("Found Vetar vendor ID: 0x%08x\n", idc);
    return 1;
  }
  printm ("wrong vendor ID. 0x%08x found, 0x%08x expected\n", idc, VETAR_VENDOR_ID);
  return 0;
}

/***************************************************************************************************************/

void f_vetar_init ()
{
  // JAM 4-2022: put here actions which are usually done in kernel module  
  int i = 0;
  unsigned char fa[4]; /* FUN0 ADER contents */
  unsigned char am = 0;
  
  // map crcsr space here:
  f_a32_vme_mas_map_CRCSR ();
  // prepare vetar crcsr space registers
  sleep(1); // JAM RIO3 TEST
  cr_vendor_id = f_set_csr_address (VME_VENDOR_ID_OFFSET);
  cr_bit_set = f_set_csr_address (BIT_SET_REG);
  cr_bit_clear = f_set_csr_address (BIT_CLR_REG);
  cr_wb_32_64 = f_set_csr_address (WB_32_64);
  cr_fun0_ader = f_set_csr_address (FUN0ADER);
  cr_fun1_ader = f_set_csr_address (FUN1ADER);

  if (f_vetar_is_present () == 0)
  {
    printm ("f_vetar_init() - Error: could not find VETAR board!\n");
    exit (-1);
  }
  /* reset the core */
  *cr_bit_set = RESET_CORE;
  usleep (10000);
  /* disable the core */
  *cr_bit_clear = ENABLE_CORE;
  /* default to 32bit WB interface */
  *cr_wb_32_64 = WB32;

  am = 0x09;    //VME_A32_USER_DATA_SCT
  fa[0] = (VETAR_REGS_ADDR >> 24) & 0xFF;
  fa[1] = (VETAR_REGS_ADDR >> 16) & 0xFF;
  fa[2] = (VETAR_REGS_ADDR >> 8) & 0xFF;
  fa[3] = (am & 0x3F) << 2;
  /* DFSR and XAM are zero */

  for (i = 0; i < 4; i++)
  {
    *(cr_fun0_ader + i) = fa[i];    // address increment is 4 here!
  }
  am = 0x39;    //VME_A24_USER_DATA_SCT;
  fa[0] = (VETAR_CTRL_ADDR >> 24) & 0xFF;
  fa[1] = (VETAR_CTRL_ADDR >> 16) & 0xFF;
  fa[2] = (VETAR_CTRL_ADDR >> 8) & 0xFF;
  fa[3] = (am & 0x3F) << 2;
  for (i = 0; i < 4; i++)
  {
    *(cr_fun1_ader + i) = fa[i];    // address increment is 4 here!
  }
  /* enable module, hence make FUN0/FUN1 available */
  *cr_bit_set = ENABLE_CORE;
  sleep (1);

  // JAM22: TODO now map control space to configure TLU and direct access:
  f_a24_vme_mas_map_CONTROL ();
  ctrl_master_control = (INTS4*) ((char*) (pl_virt_vme_base_control) + MASTER_CTRL);
  ctrl_master_data = (INTS4*) ((char*) (pl_virt_vme_base_control) + MASTER_DATA);
  ctrl_emul_dat_wd = (INTS4*) ((char*) (pl_virt_vme_base_control) + EMUL_DAT_WD);
  ctrl_window_offset_low = (INTS4*) ((char*) (pl_virt_vme_base_control) + WINDOW_OFFSET_LOW);
  ctrl_direct_access_control = (INTS4*) ((char*) (pl_virt_vme_base_control) + DIRECT_ACCESS_CONTROL);
  
  // #from vetar_wb_request// reply
  // #  <- since we cannot branch on conditions much here, always acknowledge the control register JAM22;
  printm ("first value of master control: 0x%x\n", *ctrl_master_control);
  *ctrl_master_control = 1;
  usleep (50000);
  *ctrl_master_data = 0;
  *ctrl_master_control = 3;
  //following is in probe after wishbone register:
  *ctrl_emul_dat_wd = 0;
  *ctrl_window_offset_low = 0;
  *ctrl_master_control = 0;
  printm ("first value of dactl: 0x%x\n", *ctrl_direct_access_control);
  printm ("set to direct TLU access at address 0x%x\n", tlu_address);
  *ctrl_direct_access_control = tlu_address;
  printm ("read back dactl: 0x%x\n", *ctrl_direct_access_control);
  usleep (100000);
}

/***************************************************************************************************************/

#endif // WR_USE_TLU_DIRECT_WITHOUT_KERNEL_MODULE


/***************************************************************************************************************/
